"""
    Models/Stochastic/SecModelUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ElutionModelUtils import compute_4moments
from molass_legacy.Models.ModelUtils import compute_area_props
from Experiment.ColumnTypes import get_all_poresizes
from SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.Stochastic.LognormalPoreFunc import lognormal_pore_func

DEFAULT_SIGMA = 0.1
SIGMA_LOWER_BOUND = 0.05
NUM_ENTRIES_LIST = [1000, 2000, 3000, 4000, 5000]
DEFAULT_NUM_ENTRIES = NUM_ENTRIES_LIST[1]
DEFAULT_ENTRY_TIME = 0.5

def compute_tRv(N, T, x0, m, poresize, rgs):
    rho = rgs/poresize
    rho[rho > 1] = 1
    return x0 + N*T*(1 - rho)**m

def guess_monopore_params_from_rgdist(x, y, peak_rgs, peak_trs, props, logger=None, debug=True):
    if logger is None:
        logger = logging.getLogger(__name__)
    print("guess_secparams_from_rgdist")

    N = DEFAULT_NUM_ENTRIES
    T = DEFAULT_ENTRY_TIME

    W, M1, M2, M3 = compute_4moments(x, y)

    s = np.sqrt(M2)
    init_x0 = M1 - 8*s
    me = 1.5
    mp = 1.5
    m = me + mp

    max_rg = np.max(peak_rgs)
    poresizes = np.asarray(get_all_poresizes())
    possible_poresizes = poresizes[poresizes > max_rg + 20]
    logger.info("trying for rgs=%s against possible_poresizes: %s", str(peak_rgs), str(possible_poresizes))

    dev_list = []
    for poresize in possible_poresizes:
        for N in NUM_ENTRIES_LIST:
            def x0_objective(p):
                T_, x0_ = p
                trv_ = compute_tRv(N, T_, x0_, m, poresize, peak_rgs)
                dev = np.sum((trv_ - peak_trs)**2)
                return dev
            res = minimize(x0_objective, [T, init_x0], method="Nelder-Mead")
            dev_list.append((poresize, N, *res.x, res.fun))

    dev_list = np.asarray(dev_list)
    k = np.argmin(dev_list[:,-1])

    poresize = dev_list[k,0]
    N = dev_list[k,1]
    T = dev_list[k,2]
    x0 = dev_list[k,3]
    print("poresize, N, T, x0 = (%.3g, %d, %.3g, %.3g)" % (poresize, N, T, x0))
    rho = peak_rgs/poresize
    rho[rho > 1] = 1
    niv = N*(1 - rho)**me
    tiv = T*(1 - rho)**mp
    t = x - x0

    evaluate_props = False

    def scales_objective(scales, title=None):
        cy_list = []
        for ni, ti, scale in zip(niv, tiv, scales):
            cy = scale*robust_single_pore_pdf(t, ni, ti)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        fv1 = np.sum((ty - y)**2)
        if evaluate_props:
            props_ = compute_area_props(cy_list)
            fv2 = np.sum((props_ - props)**2)
        else:
            fv2 = 0
        if title is not None:
            print(title, fv1, fv2)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y)
                for cy in cy_list:
                    ax.plot(x, cy, ":")
                ax.plot(x, ty, ":", color="red")
                fig.tight_layout()
                plt.show()
        return fv1 + fv2

    num_components = len(peak_rgs)
    init_scales = np.ones(num_components)/num_components
    if debug:
        scales_objective(init_scales, title="before minimize")
    bounds = [(0, None)]*num_components
    res = minimize(scales_objective, init_scales, bounds=bounds, method="Nelder-Mead")
    if debug:
        scales_objective(res.x, title="after minimize")
    return np.concatenate([[N, T, x0, me, mp, poresize], res.x])

def guess_lognormalpore_params_from_rgdist(x, y, peak_rgs, peak_trs, props, debug=True):
    logger = logging.getLogger(__name__)
    print("guess_lognormalpore_params_from_rgdist")
    monopore_params = guess_monopore_params_from_rgdist(x, y, peak_rgs, peak_trs, props, logger=logger, debug=debug)
    N, T, x0, me, mp, poresize = monopore_params[:6]
    init_sigma = DEFAULT_SIGMA
    init_mu = np.log(poresize)

    def mu_sigma_scales_objective(p, title=None):
        mu, sigma = p[0:2]
        cy_list = []
        for rg, scale in zip(peak_rgs, p[2:]):
            cy = lognormal_pore_func(x, scale, N, T, me, mp, mu, sigma, rg, x0)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        fv = np.sum((ty - y)**2)
        if title is not None:
            print(title, fv)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y)
                for cy in cy_list:
                    ax.plot(x, cy, ":")
                ax.plot(x, ty, ":", color="red")
                fig.tight_layout()
                plt.show()
        return fv
    
    num_components = len(peak_rgs)
    bounds = [(0, None), (SIGMA_LOWER_BOUND, None)] + [(0, None)]*num_components
    init_scales = np.ones(num_components)/num_components    # consider area
    init_params = np.concatenate([[init_mu, init_sigma], init_scales])
    if debug:
        print("init_params =", init_params)
        mu_sigma_scales_objective(init_params, title="before LNP minimize")
    res = minimize(mu_sigma_scales_objective, init_params, bounds=bounds)
    if debug:
        print("res.x =", res.x)
        mu_sigma_scales_objective(res.x, title="after LNP minimize")
    opt_params = np.concatenate([[N, T, x0, me, mp], res.x]) 
    return opt_params
                                                                        




"""
    Models.Stochastic.LognormalPoreFunc.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import lognorm
from scipy.integrate import quad_vec
from molass_legacy.SecTheory.SecPDF import FftInvPdf
from molass_legacy.KekLib.IntegrateUtils import complex_quadrature_vec
from molass_legacy.Models.Stochastic.LognormalUtils import compute_mode, compute_stdev
from molass_legacy.Models.Stochastic.ParamLimits import PORESIZE_INTEG_LIMIT

def Ksec(Rg, r, m):
    return np.power(1 - min(1, Rg/r), m)

def distr_func(r, mu, sigma):
    return lognorm.pdf(r, sigma, scale=np.exp(mu))

def integrand_impl(r, w, N, T, me, mp, mu, sigma, Rg):
    return distr_func(r, mu, sigma)*N*Ksec(Rg, r, me)*(1/(1 - w*1j*T*Ksec(Rg, r, mp)) - 1)

def lognormal_pore_cf(w, N, T, me, mp, mu, sigma, Rg, x0, const_rg_limit=False):
    if const_rg_limit:
        max_rg = PORESIZE_INTEG_LIMIT
    else:
        mode = compute_mode(mu, sigma)
        stdev = compute_stdev(mu, sigma)
        max_rg = min(PORESIZE_INTEG_LIMIT, mode + 5*stdev)

    # note that integrand_impl is a vector function because w is a vector
    integrated = complex_quadrature_vec(lambda r: integrand_impl(r, w, N, T, me, mp, mu, sigma, Rg), Rg, max_rg)[0]
    return np.exp(integrated + 1j*w*x0)     # + 1j*w*x0 may not be correct. reconsider

lognormal_pore_pdf = FftInvPdf(lognormal_pore_cf)

def lognormal_pore_func(x, scale, N, T, me, mp, mu, sigma, Rg, x0):
    return scale*lognormal_pore_pdf(x - x0, N, T, me, mp, mu, sigma, Rg, 0)  # not always the same as below
    # return scale*lognormal_pore_pdf(x, N, T, me, mp, mu, sigma, Rg, x0)

def compute_trvec(N, T, t0, me, mp, mu, sigma, rgv):
    from molass_legacy.Models.Stochastic.LognormalUtils import compute_mode, compute_stdev
    mode = compute_mode(mu, sigma)
    stdev = compute_stdev(mu, sigma)
    rmin = max(0, mode - 3*stdev)
    rmax = min(PORESIZE_INTEG_LIMIT, mode + 3*stdev)
    ones = np.ones(len(rgv))
    """
    including t0 in lambda seems to make a latent bug appear in scipy 1.11.4 or 1.12.0 in in Python 3.11.8
    excluding t0 from lambda is better, anyway
    """
    return t0 + quad_vec(lambda r: N*T*distr_func(r,mu,sigma)*np.power(1 - np.min([ones, rgv/r], axis=0), me+mp), rmin, rmax)[0]

if False:
    rv_list = []
    def distr_func_debug(r, mu, sigma):
        v = lognorm.pdf(r, sigma, scale=np.exp(mu))
        rv_list.append((r,v))
        return v

    def plot_rv_list():
        import molass_legacy.KekLib.DebugPlot as plt
        rv_array = np.array(rv_list)
        print("rv_array.shape=", rv_array.shape)
        print("r=", rv_array[:,0])
        print("v=", rv_array[:,1])
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("compute_trvec debug")
            ax.plot(*rv_array.T)
            fig.tight_layout()
            plt.show()


# coding: utf-8
"""
    Rgg.VariationalMixture.py

    based on
    https://github.com/vwrs/variational-gaussian-mixture

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
import scipy.special as sp
import matplotlib.pyplot as plt
import matplotlib as mpl

class VariationalMixture():
    """Variational Bayesian estimation of a Gaussian mixture."""

    def __init__(self, x, K):
        self.x = x
        self.N, self.D = x.shape
        self.K = K
        self.alpha0 = self.D + 1
        self.beta0 = self.D + 1
        self.nu0 = self.D + 1
        self.u0 = self._get_u0()
        self.V0 = self._get_V0()
        self._init_params()

    def _logdet(self, mat):
        (sign, logdet) = np.linalg.slogdet(mat)
        return logdet if sign == 1 else 0

    def _get_u0(self):
        return np.mean(self.x, axis=0)

    def _get_V0(self):
        diff = self.x - self.u0
        return np.linalg.inv(self.nu0 * diff.T.dot(diff) / self.N)

    def _init_u(self):
        cov = np.linalg.inv(self.nu0*self.V0) / self.beta0
        return np.random.multivariate_normal(self.u0, cov)

    def _init_params(self):
        self.alpha = np.tile(self.alpha0, self.K)
        self.beta = np.tile(self.beta0, self.K)
        self.nu = np.tile(self.nu0, self.K)
        self.u = np.array([self._init_u() for _ in range(self.K)])
        self.V = np.tile(self.V0, (self.K, 1, 1))

        self.rho = np.zeros((self.N, self.K))
        self.r = np.zeros((self.N, self.K))
        self.eta = np.zeros(self.K)

    def _update_rho(self, n, k):
        sum_digamma = np.sum([sp.digamma(0.5*(self.nu[k]+1-d))
                              for d in range(1, self.D+1)])
        diff = self.x[n, :] - self.u[k, :]

        self.rho[n, k] = np.exp(sp.digamma(self.alpha[k]) +
                                0.5*(sum_digamma +
                                     self._logdet(self.V[k, :]) -
                                     self.D/self.beta[k] - self.nu[k] *
                                     diff.dot(self.V[k, :]).dot(diff.T)
                                     )
                                )

    def _update_r(self, n, k):
        tmpsum = np.sum(self.rho[n, :])
        self.r[n, k] = self.rho[n, k] / tmpsum if tmpsum != 0 else 0

    def _update_eta(self, k):
        self.eta[k] = np.sum(self.r[:, k])

    def _update_alpha(self, k):
        self.alpha[k] = self.alpha0 + self.eta[k]

    def _update_beta(self, k):
        self.beta[k] = self.beta0 + self.eta[k]

    def _update_nu(self, k):
        self.nu[k] = self.nu0 + self.eta[k]

    def _update_u(self, k):
        self.u[k, :] = self.beta0*self.u0 + self.r[:, k].T.dot(self.x)
        self.u[k, :] /= self.beta[k]

    def _update_V(self, k):
        cov_sample = np.zeros_like(self.V0)
        diff_x = self.x - self.u[k]
        for n in range(self.N):
            cov_sample += self.r[n, k]*np.outer(diff_x[n, :], diff_x[n, :])
        diff = self.u[k] - self.u0

        self.V[k, :] = np.linalg.inv(
            np.linalg.inv(self.V0) +
            self.beta0 * np.outer(diff, diff) +
            cov_sample
        )

    def update_params(self):
        for k in range(self.K):
            for n in range(self.N):
                self._update_rho(n, k)
                self._update_r(n, k)
            self._update_eta(k)
            self._update_alpha(k)
            self._update_beta(k)
            self._update_nu(k)
            self._update_u(k)
            self._update_V(k)

    def fit(self, eps=1e-2, max_epochs=50, print_diff=False):
        epochs = 0
        while(True):
            eta_prev = self.eta.copy()
            self.update_params()
            epochs += 1
            diff = np.linalg.norm(self.eta - eta_prev)
            if print_diff:
                print(diff)
            if (epochs > 5 and
                    (diff < eps or
                     np.abs(diff_prev - diff) < eps or
                     epochs >= max_epochs)):
                break
            diff_prev = diff
        print(f'# iterations: {epochs}')
        self.map_estimate()

    def map_estimate(self):
        self.w = np.zeros(self.K)
        self.means = self.u
        self.covs = np.zeros_like(self.V)
        for k in range(self.K):
            self.w[k] = self.alpha[k]/np.ones(self.K).dot(self.alpha)
            self.covs[k, :] = np.linalg.inv(self.nu[k] * self.V[k, :])
        print(f'w: {self.w}\nmeans: {self.means}\ncovs: {self.covs}')

    def kl_divergence(self):
        sum_kl = sp.gammaln(self.K*self.alpha0 + self.N) - \
            sp.gammaln(self.K*self.alpha0)
        tmp = self.r*np.log(self.r)
        tmp[np.isnan(tmp)] = 0
        sum_kl += np.sum(tmp)
        for k in range(self.K):
            sum_kl += (sp.gammaln(self.alpha0)-sp.gammaln(self.alpha[k]) +
                       sp.multigammaln(self.nu0*.5, self.D) -
                       sp.multigammaln(self.nu[k]*.5, self.D) +
                       0.5*(self._logdet(self.V0) * self.nu0 -
                            self._logdet(self.V[k, :])*self.nu[k] +
                       (np.log(self.beta[k])-np.log(self.beta0))*self.D))
        return sum_kl

    def _make_ellipses(self, k, ax):
        cov = self.covs[k, :]

        w, v = np.linalg.eigh(cov)
        # w: 2 eigenvalues (length), v: 2 eigenvectors (rotation)
        angle = np.degrees(np.arctan2(v[0, 1], v[0, 0]))
        height, width = 2 * np.sqrt(w)  # diameter = 2 * radius

        ell = mpl.patches.Ellipse(self.means[k, :], height, width,
                                  angle,
                                  fc='lime', lw=0, alpha=.4)
        edge = mpl.patches.Ellipse(self.means[k, :], height, width,
                                   angle,
                                   fc='none', ec='forestgreen', lw=3)
        ax.add_artist(ell)
        ax.add_artist(edge)

    def plot_with_ellipses(self, figsize=(7, 7), dpi=200):
        fig, ax = plt.subplots(1, 1, figsize=figsize, dpi=dpi)
        ax.scatter(self.x[:, 0], self.x[:, 1],
                   c='deepskyblue', edgecolor='black', alpha=.5)
        for k in range(self.K):
            self._make_ellipses(k, ax)
        ax.scatter(self.means[:, 0], self.means[:, 1],
                   c='orange', edgecolor='k',
                   marker='*', s=150)
        plt.show()


def cluster_number_selection_by_kl(x, k_range, plot=False):
    kls = np.empty(len(k_range))
    for k in k_range:
        print(f'======= {k} =========')
        vgm = VariationalGaussianMixture(x, k)
        vgm.fit(print_diff=False)
        kls[k-1] = vgm.kl_divergence()
        if plot:
            vgm.plot_with_ellipses(figsize=(8, 5), dpi=100)

    argmin_k = np.argmin(kls) + 1
    print(f'argmin_k: {argmin_k}, min_kl: {kls[argmin_k]}')
    return kls, argmin_k



"""
    Models.Stochastic.BasinhoppingAnimation.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.Models.Stochastic.MomentUtils import to_moderate_props
from molass_legacy.Models.Stochastic.ParamLimits import USE_K, M1_WEIGHT, M2_WEIGHT, BASINHOPPING_SCALE, MonoporeParamsScaler
from molass_legacy.Models.Stochastic.MonoporeMoments import study_monopore_moments_impl
from molass_legacy.Models.Stochastic.MonoporeUtils import plot_monopore_curves
from SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.Stochastic.MomentsCollation import plot_collation_state_impl
from molass_legacy.Models.Stochastic.RgReliability import determine_unreliables

def basinhopping_moment_animation_impl(lrf_src,  debug=False):
    print("moments_collation_illust_impl")

    rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs(debug=False)
    peaks = lrf_src.get_peaks()
    num_peaks = len(peaks)
    model = lrf_src.model
    egh_moments_list = lrf_src.get_egh_moments_list()
    m_props = to_moderate_props(props)

    x = lrf_src.xr_x
    y = lrf_src.xr_y

    egh_cy_list = []
    for params in peaks:
        cy = model(x, params)
        egh_cy_list.append(cy)

    def plot_collation_state(ax, moments_list):
        plot_collation_state_impl(x, y, egh_cy_list, egh_moments_list, ax, moments_list)

    unreliable_indeces = determine_unreliables(peak_rgs, qualities, props, trust_max_num=None)
    nur = len(unreliable_indeces)

    use_timescale = True
    num_timescales = 1 if use_timescale else 0
    params_scaler = MonoporeParamsScaler(egh_moments_list, 0, peak_rgs, unreliable_indeces, num_timescales=num_timescales)

    temp_rgs = np.asarray(peak_rgs).copy()
    egh_moments_array = np.asarray(egh_moments_list)
    if num_peaks > 1:
        eghm1_min_diff = np.min(np.abs(np.diff(egh_moments_array[:,0])))
        if debug:
            print("eghm1_min_diff=", eghm1_min_diff)
        MIN_DIFF_PENALTY_SCALE = 1e6

    ret = study_monopore_moments_impl(lrf_src,
                                        # keep_num_components=True,     # this is required for backward compatibility
                                        # trust_max_num=2,              # do not trust Rg's exceeding this number to get possibly better fit
                                        debug=False)
    if ret is None:
        return

    monopore_params, temp_rgs, unreliable_indeces, params_scaler = ret
    N, T, t0, me, mp, poresize = monopore_params[0:6]

    init_moments_list  = study_monopore_moments_impl(lrf_src, return_init_moments=True)

    apply_diff_penalty = True
    def mnp_column_flex_objective(p, return_moments=False, debug=False):
        p = params_scaler.scale_back(p)
        N_, KT_, x0_, poresize_ = p[0:4]
        if USE_K:
            T_ = KT_/N_
        else:
            T_ = KT_
        if use_timescale:
            T_ *= p[-1]
            x0_ *= p[-1]
        if nur > 0:
            temp_rgs[unreliable_indeces] = p[4:4+nur]

        if len(temp_rgs) > 1:
            # Rg's should be in descending order
            order_penalty = max(0, np.max(np.diff(temp_rgs))) * MIN_DIFF_PENALTY_SCALE
        else:
            order_penalty = 0

        rhov = temp_rgs/poresize_
        rhov[rhov > 1] = 1
        dev_list = []
        moments_list = []
        for k, (M, rho) in enumerate(zip(egh_moments_list, rhov)):
            M1_ = x0_ + N_ * T_ * (1 - rho)**(me + mp)
            M2_ = 2 * N_ * T_**2 * (1 - rho)**(me + 2*mp)
            dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2)
            if return_moments:
                moments_list.append([M1_, M2_])
            else:
                moments_list.append(M1_)
            if debug:
                print([k], "M1, M1_ = %.3g, %.3g" % (M[0], M1_))
                print([k], "M2, M2_ = %.3g, %.3g" % (M[1], M2_))

        if return_moments:
            return moments_list

        if apply_diff_penalty and num_peaks > 1:
            m1_min_diff = np.min(np.abs(np.diff(moments_list)))
            diff_penalty = max(0, abs(1 - m1_min_diff/eghm1_min_diff) - 0.1)**2 * MIN_DIFF_PENALTY_SCALE
        else:
            diff_penalty = 0
        fv = np.sum(np.asarray(dev_list) * m_props) + diff_penalty + order_penalty   # seems better not to weight with props
        if debug:
            print("m_props=", m_props)
            print("order_penalty=", order_penalty)
            print("diff_penalty=", diff_penalty)
            print("fv=", fv)
        return fv

    if True:
        temp_params = np.asarray(study_monopore_moments_impl(lrf_src, return_init_params=True))
    else:
        temp_params = np.concatenate([[N, N*T, t0, poresize], temp_rgs[unreliable_indeces], [1]])
    temp_params_ = params_scaler.scale(temp_params)
    v = np.linspace(0, 9, 10)
    xx, yy = np.meshgrid(v, v)
    zz = np.zeros(xx.shape)
    for i in range(xx.shape[0]):
        for j in range(xx.shape[1]):
            KT_ = xx[i,j]
            t0_ = yy[i,j]
            temp_params_[[1,2]] = KT_, t0_
            zz[i,j] = mnp_column_flex_objective(temp_params_)
    
    with plt.Dp():
        fig = plt.figure(figsize=(18,5))
        ax1 = fig.add_subplot(131)
        ax2 = fig.add_subplot(132, projection="3d")
        ax3 = fig.add_subplot(133)
        fig.suptitle("Basinhopping Animation using %s" % get_in_folder(), fontsize=20)
        ax1.set_title("Animation in Elution Curves")
        plot_collation_state(ax1, init_moments_list)
        ax2.set_title("Animation in $(T,t_0,FV)$ 3D Space")
        ax2.plot_surface(xx, yy, zz)
        ax3.set_title("Animation in $(T,t_0,FV)$ Contour Map")
        ax3.contour(xx, yy, zz)
        fig.tight_layout()
        plt.show()

def basinhopping_elution_animation_impl(lrf_src,  debug=False):
    print("moments_collation_illust_impl")

    rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs(debug=False)
    peaks = lrf_src.get_peaks()
    num_peaks = len(peaks)
    model = lrf_src.model
    egh_moments_list = lrf_src.get_egh_moments_list()
    x = lrf_src.xr_x
    y = lrf_src.xr_y

    ret = study_monopore_moments_impl(lrf_src,
                                        # keep_num_components=True,     # this is required for backward compatibility
                                        # trust_max_num=2,              # do not trust Rg's exceeding this number to get possibly better fit
                                        debug=False)
    if ret is None:
        return

    monopore_params, temp_rgs, unreliable_indeces, params_scaler = ret
    N, T, t0, me, mp, poresize = monopore_params[0:6]

    rhov = temp_rgs/poresize
    rhov[rhov > 1] = 1

    def T_t0_analysis_objective(p, title=None):
        debug = title is not None
        T_ = p[0]
        t0_ = p[1]
        scales = p[2:1+num_peaks]
        cy_list = []

        for k, (rho, scale) in enumerate(zip(rhov, scales)):
            ni_ = N * (1 - rho)**me
            ti_ = T_ * (1 - rho)**mp
            cy = scale * robust_single_pore_pdf(x - t0_, ni_, ti_)
            cy_list.append(cy)
        ty = np.sum(cy_list, axis=0)
        dev = np.sum((ty - y)**2)
        return dev

    init_params = np.concatenate([[T, t0], monopore_params[6:]])

    with plt.Dp():
        fig = plt.figure(figsize=(18,5))
        ax1 = fig.add_subplot(131)
        ax2 = fig.add_subplot(132, projection="3d")
        ax3 = fig.add_subplot(133)
        fig.suptitle("Basinhopping Animation using %s" % get_in_folder(), fontsize=20)
        ax1.set_title("Animation in Elution Curves")
        ax1.plot(x, y, color="orange")
        plot_monopore_curves(ax1, x, monopore_params, temp_rgs)
        ax2.set_title("Animation in $(T,t_0,FV)$ 3D Space")
        ax3.set_title("Animation in $(T,t_0,FV)$ Contour Map")
        fig.tight_layout()
        plt.show()


"""

    CharacterFormatter.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""
from Unicode import SUPERSCRIPT, SUBSCRIPT

SUPERSCRIPT_INV = {}
for k,v in SUPERSCRIPT.items():
    SUPERSCRIPT_INV[v] = k

SUBSCRIPT_INV = {}
for k,v in SUBSCRIPT.items():
    SUBSCRIPT_INV[v] = k

def to_superscript_font(range_):
    n = range_.Characters.Count
    for i in range(1, n+1):
        c = range_.GetCharacters(i,1)
        t = SUPERSCRIPT_INV.get(c.Text)
        if t is not None:
            c.Text = t
            c.Font.Superscript = True

def to_subscript_font(range_):
    n = range_.Characters.Count
    for i in range(1, n+1):
        c = range_.GetCharacters(i,1)
        t = SUBSCRIPT_INV.get(c.Text)
        if t is not None:
            c.Text = t
            c.Font.Subscript = True

def to_italic_font(range_, targets):
    n = range_.Characters.Count
    c = range_.GetCharacters(1,n)
    text = c.Text
    for t in targets:
        i = text.find(t)
        if i >= 0:
            j = t.find('(')
            if j >= 0:
                # parentheses should not be italic
                c_ = range_.GetCharacters(i+1,j)
                c_.Font.Italic = True
                k = t.find(')')
                assert k > 1
                if t[j+1:k].isnumeric():
                    # numeric chars should not be italic
                    pass
                else:
                    c_ = range_.GetCharacters(i+j+2,k-1-j)
                    c_.Font.Italic = True
            else:
                c_ = range_.GetCharacters(i+1,len(t))
                c_.Font.Italic = True

def to_italic_subscipt_font(range_, targets):
    n = range_.Characters.Count
    c = range_.GetCharacters(1,n)
    text = c.Text
    for t in targets:
        i = text.find(t)
        if i >= 0:
            c_ = range_.GetCharacters(i+1,len(t))
            c_.Font.Italic = True
            cs = range_.GetCharacters(i+2,len(t)-1)
            cs.Font.Subscript = True



import numpy    as np
import matplotlib.pyplot as plt
import statsmodels.api      as sm

a = 2.5
b = 1.1

x = np.linspace( 0, 5, 10 )

y = a*x + b

np.random.seed( 123 )
obs = y + 0.5 * np.random.randn(10)
err = 0.3 * np.random.randn(10)
w = 1/err**2

X = sm.add_constant(x)
result = sm.WLS(obs, X, weights=w).fit()
# result = sm.OLS(obs, X).fit()
b_, a_ = result.params

print( 'result.cov_params()=',  np.diag( result.cov_params() ) )

y_ = a_*x + b_

plt.plot( x, y, color='green', label='True' )
plt.plot( x, obs, 'o', color='blue', label='Observed' )
plt.errorbar( x, obs, yerr=err, color='blue', fmt='none' )
plt.plot( x, y_, color='red', label='Fitted' )
plt.legend( loc="upper left" )
plt.show()



"""
    SecSaxs.DataSet.py

    1) successor to SerialData
    2) used only in v2 for the time being

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import logging
import inspect
import numpy as np
from bisect import bisect_right
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.ScatteringBaseUtil import apply_baseline_correction_impl
from molass_legacy.Elution.CurveUtils import get_xray_elution_vector
from .ElCurve import ElCurve
class DataSet:
    def __init__(self, xr_array, xr_ex, xr_qv, uv_array, uv_ex, uv_wv, xr_ey=None, uv_ey=None, sd=None, pre_recog=None):
        self.logger = logging.getLogger(__name__)
        self.xr_array = xr_array
        self.intensity_array = xr_array     # for compatibility with SerialData such as in OptDataSets.__init__
        self.xr_ex = xr_ex
        if xr_ey is not None:
            assert len(xr_ey) == len(xr_ex)
        self.xr_ey = xr_ey
        self.xr_qv = xr_qv
        self.qvector = xr_qv                # for backward compatibility in Optimizer.LrfExporter
        self.absorbance = None if sd is None else sd.absorbance
        self.uv_array = uv_array
        self.uv_ex = uv_ex
        if uv_ey is not None:
            assert len(uv_ey) == len(uv_ex)
        self.uv_ey = uv_ey
        self.uv_wv = uv_wv
        self.lvector = uv_wv                # for backward compatibility 
        self.sd = sd
        self.pre_recog = pre_recog
        self.orig_info = None       # for backward compatibility, or better be removed
        self.xr_curve = None
        self.uv_curve = None
        self.cd_slice = None
        if sd is None:
            self.xr_index = None
        else:
            try:
                self.xr_index = sd.xr_index
            except:
                self.xr_index = sd.xray_index

    def copy(self):
        ds = DataSet(
                self.xr_array.copy(),
                self.xr_ex.copy(),
                self.xr_qv.copy(),
                self.uv_array.copy(),
                self.uv_ex.copy(),
                self.uv_wv.copy(),
                self.xr_ey.copy(),
                self.uv_ey.copy(),
                )

        ds.absorbance = self.absorbance     # note that this is not a copy
        ds.xr_index = self.xr_index
        ds.sd = self.sd                     # necessary?
        return ds

    def get_copy(self):             # for backward compatibility
        return self.copy()

    def get_id_info(self):          # for backward compatibility
        baseline_corrected = True   # not sure
        return 'SD(id=%s, corrected=%s)' % (str(id(self)), str(baseline_corrected))

    def get_cd_slice(self):         # for backward compatibility, used in ExtrapolationSolver.py
        if self.cd_slice is None:
            qmax = bisect_right(self.qvector, get_setting('cd_eval_qmax'))
            self.cd_slice = slice(0, qmax)
        return self.cd_slice

    def get_xray_scale(self):       # for backward compatibility, used in LrfResultPool.py
        return self.xr_curve.max_y

    def get_xr_curve(self):
        if self.xr_curve is None:
            if self.sd is None:
                v1_curve = None
            else:
                v1_curve = self.sd.get_xray_curve()
            self.xr_curve = ElCurve(self.xr_ex, self.xr_ey, v1_curve=v1_curve)
        return self.xr_curve

    def get_xray_curve(self):
        # for backward compatibility with ElutionCurve
        return self.get_xr_curve()

    def get_uv_curve(self):
        if self.uv_curve is None:
            if self.sd is None:
                v1_curve = None
            else:            
                v1_curve = self.sd.get_uv_curve()
            self.uv_curve = ElCurve(self.uv_ex, self.uv_ey, v1_curve=v1_curve)
        return self.uv_curve

    def get_xr_data_separate_ly(self):
        D = self.xr_array[:,:,1].T
        E = self.xr_array[:,:,2].T
        xr_curve = self.get_xr_curve()
        return D, E, self.xr_qv, xr_curve

    def get_uv_data_separate_ly(self):
        D = self.uv_array
        uv_curve = self.get_uv_curve()
        return D, None, self.uv_wv, uv_curve

    def apply_baseline_correction( self, mapped_info, basic_lpm=False, progress_cb=None, return_base=False, debug_obj=None  ):
        frame = inspect.stack()[1]
        self.logger.info("apply baseline_correction called from %s(%d) with basic_lpm=%s", frame.filename, frame.lineno, str(basic_lpm))
        self.logger.info("id:%s, code_context:%s", str(id(self)),''.join(frame.code_context)[:-1])

        debug = False
        if debug:
            xr_array_copy = self.xr_array.copy()

        jvector = np.arange(self.xr_array.shape[0])
        base = apply_baseline_correction_impl(
            jvector, self.qvector, self.xr_array,
            mapped_info, basic_lpm=basic_lpm,
            index=self.xr_index,
            ecurve=self.get_xr_curve(),
            progress_cb=progress_cb,
            return_base=return_base, logger=self.logger, debug_obj=debug_obj )

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.DataStructure.MatrixData import simple_plot_3d
            qv = self.qvector

            with plt.Dp():
                ey1 = get_xray_elution_vector(qv, xr_array_copy)[0]
                ey2 = get_xray_elution_vector(qv, self.xr_array)[0]
                fig, ax = plt.subplots()
                ax.set_title("apply_baseline_correction: elution curves")
                ax.plot(ey1, label="original")
                ax.plot(ey2, label="corrected")
                ax.legend()
                fig.tight_layout()
                plt.show()

            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5), subplot_kw=dict(projection="3d"))
                fig.suptitle("apply_baseline_correction: 3d views")
                simple_plot_3d(ax1, self.xr_array[:,:,1].T, x=qv)
                simple_plot_3d(ax2, self.intensity_array[:,:,1].T, x=qv)
                fig.tight_layout()
                plt.show()

        try:
            self.set_xr_dependents(debug=debug)
        except:
            # not fatal as in the degraded case of 20161124
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "apply_baseline_correction: ")

        return base

    def get_xray_elution_vector( self ):
        xr_ey, slice_ = get_xray_elution_vector(self.qvector, self.xr_array)
        self.xr_slice = slice_
        self.xr_index = (slice_.start + slice_.stop)//2
        self.xray_index = self.xr_index     # for backward compatibility with LrfSolver.py
        return xr_ey

    def set_xr_dependents(self, debug=False):
        from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve

        self.xr_ey = self.get_xray_elution_vector()

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("set_ivector_etcetera")
                ax.plot(self.ivector, label="ivector")
                ax.legend()
                fig.tight_layout()
                plt.show()

        # self.xr_curve = ElCurve(self.xr_ex, self.xr_ey)   # ElCurve is not mature enough to do this
        self.xr_curve = ElutionCurve(self.xr_ey, x=self.xr_ex)

    def _get_analysis_copy_impl(self, given_pre_recog):
        xr_array = self.xr_array
        xr_ex = self.xr_ex
        xr_ey = self.xr_ey
        xr_qv = self.xr_qv
        uv_array = self.uv_array
        uv_ex = self.uv_ex
        uv_ey = self.uv_ey
        uv_wv = self.uv_wv

        kwargs = {}

        xr_restrict_list = get_setting('xr_restrict_list')
        if xr_restrict_list is None:
            xr_array_ = xr_array
            xr_ex_ = xr_ex
            xr_ey_ = xr_ey
            xr_qv_ = xr_qv
        else:
            # print('xr_restrict_list=', xr_restrict_list)
            elution_restrict = xr_restrict_list[0]
            if elution_restrict is None:
                eslice = slice(None, None)
            else:
                eslice = slice(elution_restrict.start, elution_restrict.stop)
            angle_restrict = xr_restrict_list[1]
            if angle_restrict is None:
                sslice = slice(None, None)
            else:
                sslice = slice(angle_restrict.start, angle_restrict.stop)
            xr_array_ = xr_array[eslice,sslice,:]
            xr_ex_ = xr_ex[eslice]
            xr_qv_ = xr_qv[sslice]
            self.logger.info('xr_array has been restricted from %s to %s' % (str(xr_array.shape), str(xr_array_.shape)))

        if xr_ey is not None:
            kwargs['xr_ey'] = xr_ey[eslice].copy()

        uv_restrict_list = get_setting('uv_restrict_list')
        if uv_restrict_list is None:
            uv_array_ = uv_array
            uv_ex_ = uv_ex
            uv_ey_ = uv_ey
            uv_wv_ = uv_wv
        else:
            elution_restrict = uv_restrict_list[0]
            if elution_restrict is None:
                eslice = slice(None, None)
            else:
                eslice = slice(elution_restrict.start, elution_restrict.stop)
            wl_restrict = uv_restrict_list[1]
            if wl_restrict is None:
                wslice = slice(None, None)
            else:
                wslice = slice(wl_restrict.start, wl_restrict.stop+1)
            uv_array_ = uv_array[wslice, eslice]
            uv_ex_ = uv_ex[eslice]
            uv_wv_ = uv_wv[wslice]
            self.logger.info('uv_array has been restricted from %s to %s' % (str(uv_array.shape), str(uv_array_.shape)))

        if uv_ey is not None:
            kwargs['uv_ey'] = uv_ey[eslice].copy()

        # return DataSet(xr_array, xr_ex, xr_qv, uv_array, uv_ex, uv_wv, xr_ey=xr_ey, uv_ey=uv_ey, sd=analysis_sd)
        analysis_copy = DataSet(xr_array_.copy(),
                                xr_ex_.copy(),
                                xr_qv_.copy(),
                                uv_array_.copy(),
                                uv_ex_.copy(),
                                uv_wv_.copy(),
                                **kwargs
                                )

        self.logger.info("sd analysis copy has been made with trimming info UV %s and Xray %s", str(uv_restrict_list), str(xr_restrict_list))
        self.logger.info('sd analysis copy %s made with xray_array.shape=%s from %s', analysis_copy.get_id_info(), str(analysis_copy.intensity_array.shape), analysis_copy.get_id_info() )

        return analysis_copy

def get_slices_from_list(trim_list):
    slice_list = []
    if trim_list is None:
        slice_list = [slice(None, None)] * 2
    else:
        for trim_info in trim_list:
            if trim_info is None:
                start = None
                stop = None
            else:
                start = trim_info.start
                stop = trim_info.stop
            slice_list.append(slice(start, stop))
    return slice_list

def get_vector_from_slice(size, slice_):
    start = slice_.start
    if start is None:
        start = 0
    stop = slice_.stop
    if stop is None:
        stop = size
    return np.arange(start, stop)

def copy_create_dataset_from_sd(original_sd, analysis_sd, debug=False):
    # task: reconsider this procedure        
    if False:
        xr_restrict_list = get_setting("xr_restrict_list")
        uv_restrict_list = get_setting("uv_restrict_list")

        xr_eslice, xr_aslice = get_slices_from_list(xr_restrict_list)
        uv_eslice, uv_wslice = get_slices_from_list(uv_restrict_list)

        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig, (ax1,ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("copy_create_dataset_from_sd")
            for k, (ax, title, sd) in enumerate([(ax1, "original_sd", original_sd),(ax2, "analysis_sd", analysis_sd)]):
                ax.set_title(title)
                xr_curve = sd.get_xr_curve()
                ax.plot(xr_curve.x, xr_curve.y)
                if k == 0:
                    for j in xr_eslice.start, xr_eslice.stop:
                        if j is None:
                            continue
                        print("j=", j, xr_curve.x[[0,-1]])
                        ax.axvline(j, color="red")
            fig.tight_layout()
            plt.show()

    xr_array = analysis_sd.intensity_array.copy()
    # xr_ex = get_vector_from_slice(original_sd.intensity_array.shape[0], xr_eslice)
    try:
        xr_ex = analysis_sd.xr_ex.copy()
        xr_ey = analysis_sd.xr_ey.copy()
        xr_qv = analysis_sd.xr_qv.copy()
    except:
        xr_curve = analysis_sd.get_xr_curve()
        xr_ex = xr_curve.x.copy()
        xr_ey = xr_curve.y.copy()
        assert len(xr_ex) == len(xr_ey)
        xr_qv = analysis_sd.qvector.copy()

    try:
        uv_array = analysis_sd.uv_array.copy()
        uv_ex = analysis_sd.uv_ex.copy()
        uv_ey = analysis_sd.uv_ey.copy()
        uv_wv = analysis_sd.uv_wv.copy()
    except:
        uv_array = analysis_sd.conc_array.copy()
        # uv_ex = get_vector_from_slice(original_sd.conc_array.shape[1], uv_eslice)
        # uv_ey = analysis_sd.absorbance.a_vector
        uv_curve = analysis_sd.get_uv_curve()
        uv_ex = uv_curve.x.copy()
        uv_ey = uv_curve.y.copy()
        assert len(uv_ex) == len(uv_ey)
        uv_wv = analysis_sd.lvector.copy()

    return DataSet(xr_array, xr_ex, xr_qv, uv_array, uv_ex, uv_wv, xr_ey=xr_ey, uv_ey=uv_ey, sd=analysis_sd)




"""
Computing the 3D Radial Distribution Function from Particle Positions: An Advanced Analytic Approach
Bernd A. F. Kopera and Markus Retsch
Analytical Chemistry 2018 90 (23), 13909-13914
DOI: 10.1021/acs.analchem.8b03157
"""

from numpy import sqrt
from random import random
from numpy . random import normal
from matplotlib.pyplot import *
from mpl_toolkits.mplot3d import Axes3D

"""
This program calculates FCC coordinates.
A 3D Gaussian disturbance can be added.
Random Sphere Point Picking Algorithm from:
G. Marsaglia, The Annals of Mathematical Statistics,
1972, Vol. 43, No. 2, 845-646
"""

### DASHBOARD

SimuName = "FCC_Blurr_Cube" # Name of the Experiment and the Files
GitterKonst = sqrt(2)       # Cell constant for the FCC unit cell
PartDiam = 1                # Diameter of the Particles
Blurr = True                # add a 3D gaussian blurr?
sigma = 0.07                # standard deviation of the blurr
Lx = 4                      # Distance in x-direction from -Lx to Lx
Ly = 4                      # Distance in y-direction from -Ly to Ly
Lz = 4                      # Distance in z-direction from -Lz to Lz

"""***********************************************************"""

def calcFCCCoor(GitterKonst, Lx, Ly, Lz):

    CoorList = []

    XLim = int(Lx/GitterKonst*5)
    YLim = int(Ly/GitterKonst*5)
    ZLim = int(Lz/GitterKonst*5)

    L = GitterKonst/2

    A = [0, L, L]
    B = [L, 0, L]
    C = [L, L, 0]

    for h in range(-ZLim*2, ZLim*2, 1):
        for k in range(-YLim*2, YLim*2, 1):
            for l in range(-XLim*2, XLim*2, 1):

                NewX = h * A[0] + k * B[0] + l * C[0]
                NewY = h * A[1] + k * B[1] + l * C[1]
                NewZ = h * A[2] + k * B[2] + l * C[2]

                if (abs(NewX) < Lx and
                    abs(NewY) < Ly and
                    abs(NewZ) < Lz):

                    CoorList.append([NewX, NewY, NewZ])

    return CoorList

"""***********************************************************"""

def getRandOffset(sigma):

    x1 = 2
    x2 = 2

    while x1**2 + x2**2 >= 1:

        x1 = random()*2-1
        x2 = random()*2-1

    R = normal(0, sigma)

    Wurzel = sqrt(1 - x1**2 - x2**2)
    x = 2 * x1 * Wurzel
    y = 2 * x2 * Wurzel
    z = 1 - 2 * (x1**2 + x2**2)

    return [x * R, y * R, z * R]

"""***********************************************************"""

def addGaussianBlurr(CoorList, sigma):

    BlurrList = []

    for P in CoorList:

        Offset = getRandOffset(sigma)

        NewX = P[0] + Offset[0]
        NewY = P[1] + Offset[1]
        NewZ = P[2] + Offset[2]

        BlurrList.append([NewX, NewY, NewZ])

    return BlurrList

"""***********************************************************"""

def plotCorr(CoorList, SimuName):

    X = [0] * len(CoorList)
    Y = [0] * len(CoorList)
    Z = [0] * len(CoorList)

    for P in range(0, len(CoorList), 1):

        X[P] = CoorList[P][0]
        Y[P] = CoorList[P][1]
        Z[P] = CoorList[P][2]

    fig = figure()
    # ax = fig.add_subplot(111, projection = "3d", aspect="equal") 
    # NotImplementedError: Axes3D currently only supports the aspect argument 'auto'. You passed in 'equal'.
    ax = fig.add_subplot(111, projection = "3d")

    ax.scatter(X, Y, Z)

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    savefig(SimuName + "_3D.png", dpi = 600)
    show ( )

    return True

"""***********************************************************"""

def saveData(CoorList, SimuName):

    DataFile = open(SimuName + "_rawCoor.txt", "w")

    for Particle in CoorList:

        DataFile.write( str(Particle[0]) + "\t" +
                        str(Particle[1]) + "\t" +
                        str(Particle[2]) + "\n")

    DataFile.close()

    return True

"""***********************************************************"""

def createAutoCADFile(CoorList, PartDiam, SimuName):

    CADFile = open(SimuName + "_CAD.scr", "w")

    for Particle in CoorList:

        CADFile.write("SPHERE\n" ) # SPHERE command
        CADFile.write(str(round(Particle[0], 5)) + ",")     # X-Position
        CADFile.write(str(round(Particle[1], 5)) + ",")     # Y-Position
        CADFile.write(str(round(Particle[2], 5)) + "\n")    # Z-Position
        CADFile.write(str(PartDiam/2) + "\n")   # Radius

    CADFile.write("\n")     # Blank line at the end
    CADFile.close()

    return True

"""***********************************************************"""

def generate_fcc_coords(blurr=True):
    import numpy as np
    print("Calculating FCC Coordinates.")
    FCC = calcFCCCoor(GitterKonst, Lx, Ly, Lz)
    if blurr:
        print("Adding a Gaussian blurr with Sigma = " + str(sigma))
        FCC = addGaussianBlurr(FCC, sigma)
    return np.array(FCC)

###MAIN
if __name__ == '__main__':

    print("Calculating FCC Coordinates.")
    FCC = calcFCCCoor(GitterKonst, Lx, Ly, Lz)

    print( "Particle Number : " +str(len(FCC)))

    if Blurr == True :

        print("Adding a Gaussian blurr with Sigma = " + str(sigma))
        FCC = addGaussianBlurr(FCC, sigma)

    print("Saveing the Coordinates to a .txt file.")
    saveData (FCC, SimuName )

    print("Creating a AutoCAD file.")
    createAutoCADFile (FCC, PartDiam , SimuName)

    print( "Plotting the Coordinates.")
    plotCorr(FCC, SimuName)



"""
    Models.Stochastic.PairedAnalysisImpl.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Stochastic.DispersiveUtils import NUM_SDMCUV_PARAMS

def paired_analysis_impl(names, num_components, info_list, rg_info_list, unreliable_indeces,
                         sdm_params_info_list, mapping_list, mapping, moments_list, exec_spec
                         ):
    import SSDC.SsdcAnalysisUtils
    reload(SSDC.SsdcAnalysisUtils)
    from SSDC.SsdcAnalysisUtils import plot_ssdc_result
    print("paired_analysis_impl")

    for k, (sdm_params, temp_rgs, temp_bouns) in enumerate(sdm_params_info_list):
        print([k], sdm_params)
        print([k], temp_rgs)
        print([k], temp_bouns)
        N, K, x0, poresize, N0, tI = sdm_params[0:NUM_SDMCUV_PARAMS]
        print([k], poresize, N0, tI)

    thick_params, thick_rgs, thick_bounds = sdm_params_info_list[0]
    thick_N, thick_K, thick_x0, thick_poresize, thick_N0, thick_tI = thick_params[0:NUM_SDMCUV_PARAMS]
    thick_T = thick_K/thick_N

    thin_params, thin_rgs, thin_bounds = sdm_params_info_list[1]
    thin_N, thin_K, thin_x0, thin_poresize, thin_N0, thin_tI = thin_params[0:NUM_SDMCUV_PARAMS]
    thin_T = thin_K/thin_N


    init_rgs = np.average([thick_rgs, thin_rgs], axis=0)
    temp_rgs = init_rgs.copy()
    print("temp_rgs=", temp_rgs)
    print("mapping=", mapping)
    a, b = mapping

    me = 1.5
    mp = 1.5

    thick_M1_tv = [M[0] for M in moments_list[0]]
    thick_M2_tv = [np.sqrt(M[1]) for M in moments_list[0]]
    thin_M1_tv  = [M[0] for M in moments_list[1]]
    thin_M2_tv  = [np.sqrt(M[1]) for M in moments_list[1]]

    minimal = False
    use_basinhoppng = False
    if minimal:
        unreliable_indeces = np.array([], dtype=int)
    else:
        unreliable_indeces = np.array(unreliable_indeces, dtype=int)

    def paired_objective(p):
        poresize = p[0]
        if minimal:
            thick_tI_, thin_N0_ = p[1:3]
            thick_N_ = thick_N
            thick_T_ = thick_T
            thin_N_ = thin_N
            thin_T_ = thin_T
            thick_x0_ = thick_x0
            thin_x0_ = thin_x0
        else:
            thick_N_, thick_K_, thick_x0_, thick_tI_ = p[1:5]
            thick_T_ = thick_K_/thick_N_
            thin_N_, thin_K_, thin_x0_, thin_N0_ = p[5:9]
            thin_T_ = thin_K_/thin_N_
        temp_rgs[unreliable_indeces] = p[9:]

        rhov = temp_rgs/poresize
        rhov[rhov > 1] = 1
        thin_tI_ = thick_tI_*a + b
        thick_M1v = thick_x0_ + thick_N_*thick_T_*(1 - rhov)**(me+mp)
        thin_M1v  =  thin_x0_ +  thin_N_* thin_T_*(1 - rhov)**(me+mp)
        thick_M2v = np.sqrt(2*thick_N_*thick_T_**2*(1 - rhov)**(me+2*mp) + (thick_M1v - thick_tI_)**2/thick_N0)
        thin_M2v  = np.sqrt(2* thin_N_* thin_T_**2*(1 - rhov)**(me+2*mp) + (thin_M1v  - thin_tI_ )**2/ thin_N0_)
        return (  np.log(np.sum((thick_M1v - thick_M1_tv)**2)) + np.log(np.sum((thick_M2v - thick_M2_tv)**2))
                + np.log(np.sum(( thin_M1v -  thin_M1_tv)**2)) + np.log(np.sum(( thin_M2v -  thin_M2_tv)**2))
                )
    init_poresize = (thick_poresize + thin_poresize)/2
    poresize_bounds = exec_spec["poresize_bounds"]
    if minimal:
        init_params = np.array([init_poresize, thick_tI, thin_N0])
        bounds = [poresize_bounds, (-3000, 0), (3600, thick_N0)]
    else:
        init_params = np.concatenate([[init_poresize,
                                    thick_N, thick_K, thick_x0, thick_tI,
                                     thin_N,  thin_K,  thin_x0, thin_N0],
                                    init_rgs[unreliable_indeces]])
        print("thick_bounds[4]=", thick_bounds[4])
        bounds = [poresize_bounds,
                (300, 3000), (300, 2000), (thick_x0 - 100, thick_x0 + 100),
                thick_bounds[4],    # better widen?
                (300, 3000), (300, 2000), (thin_x0 - 100, thin_x0 + 100), (900, thick_N0)] + [(10, 100)]*len(unreliable_indeces)

    if use_basinhoppng:
        minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds)
        res = basinhopping(paired_objective, init_params, minimizer_kwargs=minimizer_kwargs)        
    else:
        res = minimize(paired_objective, init_params, method='Nelder-Mead', bounds=bounds)

    p = res.x
    poresize = p[0]
    if minimal:
        thick_tI, thin_N0 = p[1:3]
    else:
        thick_N, thick_K, thick_x0, thick_tI = p[1:5]
        thin_N, thin_K, thin_x0, thin_N0 = p[5:9]
        # temp_rgs[unreliable_indeces] = p[9:]
    thin_tI = thick_tI*a + b
    opt_thick_params = np.concatenate([[thick_N, thick_K, thick_x0, poresize, thick_N0, thick_tI], thick_params[NUM_SDMCUV_PARAMS:]])
    opt_thin_params  = np.concatenate([[thin_N,  thin_K,  thin_x0,  poresize,  thin_N0,  thin_tI],  thin_params[NUM_SDMCUV_PARAMS:]])
    print(thick_params)
    print(opt_thick_params)
    print(thin_params)
    print(opt_thin_params)
    opt_sdm_params_info_list = [(opt_thick_params, temp_rgs), (opt_thin_params, temp_rgs)]
    plot_ssdc_result(names, info_list, rg_info_list, unreliable_indeces,
                     opt_sdm_params_info_list, mapping_list, mapping)


"""

    GuiDenssTools.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy._MOLASS.SerialSettings import get_setting

class GuiDenssToolsMenu(Tk.Menu):
    def __init__(self, parent, menubar ):
        self.parent = parent

        Tk.Menu.__init__(self, menubar, tearoff=0 )
        menubar.add_cascade( label="DENSS Tools", menu=self )
        self.add_command( label="DENSS GUI", command=self.show_denss_dialog )
        self.add_command( label="DENSS Manager", command=self.show_denss_manager )
        self.add_command( label="Electron Density Viewer", command=self.show_ed_viewer )
        self.add_command( label="SAXS Simulator", command=self.show_saxs_simulator )
        self.update_states()

    def update_states(self):
        from molass_legacy.Env.EnvInfo import get_global_env_info
        analysis_folder = get_setting('analysis_folder')
        env_info = get_global_env_info()
        state = Tk.DISABLED if analysis_folder is None or env_info is None else Tk.NORMAL

    def show_denss_dialog( self ):
        from molass_legacy.DENSS.DenssGui import DenssGuiDialog
        analysis_folder = get_setting("analysis_folder")
        if analysis_folder is None:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            yn = MessageBox.askyesno("Output Folder Preference Query",
                    '"Analysis Result Folder" is not yet specified.\n'
                    "You can continue without it, but you have to specify the output folder each time.\n"
                    "Or, if you specify it in advance, successive output folders will be automatically created as shown below.\n"
                    "    .../analysis-nnn/DENSS/000\n"
                    "    .../analysis-nnn/DENSS/001\n"
                    "          ︙\n"
                    "Do you wish to continue simply (i.e., without auto-creating)?",
                    parent=self.parent)
            if not yn:
                return

        self.parent.update_analysis_folder()
        dialog = DenssGuiDialog(self.parent)
        dialog.show()

    def show_denss_manager(self):
        print('show_denss_manager')
        from molass_legacy.DENSS.DenssManagerDialog import show_manager_dialog
        show_manager_dialog(self.parent)

    def show_ed_viewer(self):
        print('show_ed_viewer')
        from molass_legacy.Saxs.EdViewer import EdViewer
        from molass_legacy.KekLib.OurMatplotlib import reset_to_default_style
        viewer = EdViewer(self.parent)
        viewer.show()
        reset_to_default_style()

    def show_saxs_simulator(self):
        print('show_saxs_simulator')
        from molass_legacy.Saxs.SaxsSimulator import SaxsSimulator
        from molass_legacy.KekLib.OurMatplotlib import reset_to_default_style
        simulator = SaxsSimulator(self.parent)
        simulator.show()
        reset_to_default_style()



"""
    Optimizer.SvdTutorial.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class SvdTutorial(Dialog):
    def __init__(self, parent, dialog):
        self.parent = parent
        self.dialog = dialog

        Dialog.__init__(self, parent, "SVD Tutorial", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig = plt.figure(figsize=(18,11))

        self.fig = fig
        self.draw()
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.draw()

    def buttonbox(self):
        lower_frame = Tk.Frame(self)
        lower_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(lower_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(lower_frame, width=padx)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(lower_frame)
        box.pack(side=Tk.RIGHT)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=padx, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw(self):
        fig = self.fig
        gs = GridSpec(2,4)

        axes = []
        for i in range(2):
            axes_row = []
            for j in range(4):
                ax = fig.add_subplot(gs[i,j], projection="3d")
                axes_row.append(ax)
            axes.append(axes_row)

        fig.tight_layout()



"""
    Models/RateTheory/RobustEDM_Utils.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from matplotlib.widgets import Slider
import molass_legacy.KekLib.DebugPlot as plt
from SecTheory.Edm import edm_func
from .RobustEDM import edm_impl
from molass_legacy.Peaks.MomentsUtils import compute_moments      # for backward compatibility

DRIVE ="D:"
DATA_FOLDER = DRIVE + r"\PyTools\Data"
SAVE_FOLDER = DRIVE + r"\TODO\20230807\training_data"

def guess_init_params(M):
    t0 = M[1]/2
    u = 30/M[1]*3
    a = 0.2
    b = 0.2
    e = 0.4
    Dz = 0.02
    cinj = M[0]/2.0 * 0.2
    return np.array([t0, u, a, b, e, Dz, cinj])

def save_training_data():
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    print("save_training_data")

    data_list = [
        r"\20161104\BL-6A\AhRR",
        r"\20161104\BL-10C\OA",
        r"\20161104\BL-10C\Ald",
        r"\20161216\Backsub",
        r"\20170512\Backsub",
        # r"\20171203",
        r"\20180526\GI",
        r"\20180526\OA",
        r"\20180602",
        r"\20190309_1",
        ]

    sp = StandardProcedure()

    for k, node in enumerate(data_list):
        in_folder = DATA_FOLDER + node
        print([k], node)
        sd = sp.load_old_way(in_folder)
        xr_ecurve = sd.get_xray_curve()
        uv_ecurve = sd.get_uv_curve()

        for type_, ecurve in ('xr', xr_ecurve),('uv', uv_ecurve):
            x = ecurve.x
            y = ecurve.y
            path = r"%s\%04d-%s.dat" % (SAVE_FOLDER, k, type_)
            np.savetxt(path, np.array([x, y]).T)

def try_optimize(x, y, init_params, debug=True):

    def objective(p):
        y_ = edm_impl(x, *p)
        return np.sum((y_ - y)**2)

    ret = minimize(objective, init_params)
    score = np.sqrt(ret.fun)/np.max(y)

    if debug:
        print("init_params=", init_params)
        print("ret.x=", ret.x)
        print("score=", score)


        def add_sliders(fig, line, slider_axes, sliders, left, width, slider_params):
            slider_specs = [    ("t0", -50, 200, slider_params[0]),
                                ("u", 0, 2, slider_params[1]),
                                ("a", 0, 2, slider_params[3]),
                                ("b", -4, 1, slider_params[3]),
                                ("e", 0, 2, slider_params[4]),
                                ("Dz", 0, 1, slider_params[5]),
                                ("cinj", 0, 3, slider_params[6]),
                                # ("tinj", 0, 10, slider_params[7]),
                                ]

            def slider_update(k, val):
                # print([k], "slider_update", val)
                init_params[k] = val
                y_ = edm_impl(x, *init_params)
                line.set_data(x, y_)
                dp = plt.get_dp()
                dp.draw()

            for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
                ax_ = fig.add_axes([left, 0.8 - 0.08*k, width, 0.03])
                slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
                slider.on_changed(lambda val, k_=k: slider_update(k_, val))
                slider_axes.append(ax_)
                sliders.append(slider)

        with plt.Dp():
            fig, (ax, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            ax.plot(x, y)

            edm_line1, = ax.plot(x, edm_impl(x, *init_params))
            ax2.plot(x, y)
            edm_line2, = ax2.plot(x, edm_impl(x, *ret.x))

            slider_axes = []
            sliders = []
            add_sliders(fig, edm_line1, slider_axes, sliders, 0.15, 0.25, init_params)
            add_sliders(fig, edm_line2, slider_axes, sliders, 0.65, 0.25, ret.x)

            fig.tight_layout()
            plt.show()

    return ret.x, score

def train():
    from glob import glob
    fh = open("moments-params.csv", "w")

    for file in glob(SAVE_FOLDER + r"\*.dat"):
        print(file)
        x, y = np.loadtxt(file).T
        M = compute_moments(x, y)
        init_params = guess_init_params(M)
        params, score = try_optimize(x, y, init_params)
        if score < 1:
            fh.write(",".join(["%.3g" % v for v in [*M, *params]]) + "\n")

    fh.close()



"""
    PyChemLib2D.py

    functions borrowed from
    https://pythoninchemistry.org/sim_and_scat/intro

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from scipy.constants import Boltzmann
mass_of_argon = 39.948 # amu

from .PyChemLib import lj_force

NDIM = 2

def init_velocity(T, num_particles, mass):
    """
    Initialise the velocities for a series of 
    particles.
    
    Parameters
    ----------
    T: float
        Temperature of the system at 
        initialisation (K)
    number_of_particles: int
        Number of particles in the system
    
    Returns
    -------
    ndarray of floats
        Initial velocities for a series of 
        particles (eVs/Åamu)
    """
    R = np.random.rand(num_particles, NDIM) - 0.5
    return R * np.sqrt(Boltzmann * T / (
        mass * 1.602e-19))

def get_accelerations(positions, mass, epsilon, sigma):
    """
    Calculate the acceleration on each 
    particle as a  result of each other 
    particle. 
    N.B. We use the Python convention of
    numbering from 0.
    
    Parameters
    ----------
    positions: ndarray of floats
        The positions, in a single dimension, 
        for all of the particles (Å)
        
    Returns
    -------
    ndarray of floats
        The acceleration on each particle (eV/Åamu)
    """
    n = positions.shape[0]
    dim = positions.shape[1]
    assert dim == NDIM
    accel_x = np.zeros((n, n, dim))
    for i in range(0, n - 1):
        for j in range(i + 1, n):
            r_x = positions[j] - positions[i]
            rmag = np.sqrt(np.sum(r_x**2))
            force_scalar = lj_force(rmag, epsilon, sigma)
            force_x = force_scalar * r_x / rmag
            accel_x[i, j] = force_x / mass #eV Å-1 amu-1
            # appling Newton's third law
            accel_x[j, i] = - force_x / mass
    return np.sum(accel_x, axis=0)

def demo_accel():
    mass_of_argon = 39.948 # amu
    epsilon = 0.0103
    sigma = 3.4
    accel = get_accelerations(np.array([(1,1), (5,5), (10,10)]), mass_of_argon, epsilon, sigma)
    print('Acceleration on particle 0 = %.3e, %.3e eV/Åamu' % tuple(accel[0]))
    print('Acceleration on particle 1 = %.3e, %.3e eV/Åamu' % tuple(accel[1]))
    print('Acceleration on particle 2 = %.3e, %.3e eV/Åamu' % tuple(accel[2]))

def update_pos(x, v, a, dt, box_limits):
    """
    Update the particle positions.
    
    Parameters
    ----------
    x: ndarray of floats
        The positions of the particles in a 
        single dimension
    v: ndarray of floats
        The velocities of the particles in a 
        single dimension
    a: ndarray of floats
        The accelerations of the particles in a 
        single dimension
    dt: float
        The timestep length
    
    Returns
    -------
    ndarray of floats:
        New positions of the particles in a single 
        dimension
    """
    zzzz = np.zeros(len(x))
    xmin = np.ones(len(x))*box_limits[0][0]
    xmax = np.ones(len(x))*box_limits[0][1]
    ymin = np.ones(len(x))*box_limits[1][0]
    ymax = np.ones(len(x))*box_limits[1][1]
    x_ = x + v * dt + 0.5 * a * dt * dt
    exmin = np.min([x_[:,0] - xmin, zzzz], axis=0)
    exmax = np.max([x_[:,0] - xmax, zzzz], axis=0)
    eymin = np.min([x_[:,1] - ymin, zzzz], axis=0)
    eymax = np.max([x_[:,1] - ymax, zzzz], axis=0)
    ex = exmin + exmax
    ey = eymin + eymax
    e_ = np.array([ex, ey]).T
    v[ex!=0,0] *= -1
    v[ey!=0,1] *= -1
    return x_ - 2*e_

def update_velo(v, a, a1, dt):
    """
    Update the particle velocities.
    
    Parameters
    ----------
    v: ndarray of floats
        The velocities of the particles in a 
        single dimension (eVs/Åamu)
    a: ndarray of floats
        The accelerations of the particles in a 
        single dimension at the previous 
        timestep (eV/Åamu)
    a1: ndarray of floats
        The accelerations of the particles in a
        single dimension at the current 
        timestep (eV/Åamu)
    dt: float
        The timestep length
    
    Returns
    -------
    ndarray of floats:
        New velocities of the particles in a
        single dimension (eVs/Åamu)
    """
    return v + 0.5 * (a + a1) * dt

def run_md(num_particles, dt, number_of_steps, initial_temp, x, mass, box_limits, epsilon, sigma):
    """
    Run a MD simulation.
    
    Parameters
    ----------
    dt: float
        The timestep length (s)
    number_of_steps: int
        Number of iterations in the simulation
    initial_temp: float
        Temperature of the system at 
        initialisation (K)
    x: ndarray of floats
        The initial positions of the particles in a 
        single dimension (Å)
        
    Returns
    -------
    ndarray of floats
        The positions for all of the particles 
        throughout the simulation (Å)
    """
    positions = np.zeros((number_of_steps, num_particles, NDIM))
    v = init_velocity(initial_temp, num_particles, mass)
    a = get_accelerations(x, mass, epsilon, sigma)
    for i in range(number_of_steps):
        x = update_pos(x, v, a, dt, box_limits)
        a1 = get_accelerations(x, mass, epsilon, sigma)
        v = update_velo(v, a, a1, dt)
        a = np.array(a1)
        positions[i,:] = x
    return positions

def demo_run_md():
    epsilon = 0.0103
    sigma = 3.4

    box_limits = np.array([(-10, 20), (-10, 20)])
    x = np.array([(0,0), (5,5), (5,0), (0, 5), (10,10), (5,-9), (-9,5)])
    num_particles = len(x)
    sim_pos = run_md(num_particles, 0.1, 10000, 300, x, mass_of_argon, box_limits, epsilon, sigma)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))

    r = np.linspace(3.5, 8, 100)
    ax1.set_title("Lennard-Jones Potential")
    ax1.plot(r, lj_force(r, epsilon, sigma))
    ax1.set_xlabel("Distance")
    ax1.set_ylabel("Potential")

    ax2.set_title("Particle Movement Animation")
    point_arts, = ax2.plot(sim_pos[0,:,0], sim_pos[0,:,1], "o")
    ax2.set_xlabel(r'X')
    ax2.set_ylabel(r'Y')
    ax2.set_xlim(-10, 20)
    ax2.set_ylim(-10, 20)
    # ax2.legend()
    fig.tight_layout()

    def init():
        return point_arts,

    def update(i):
        pos = sim_pos[i*100]
        point_arts.set_data(pos[:,0], pos[:,1])
        return point_arts,

    ani = FuncAnimation(fig, update, frames=len(sim_pos)//100,
                        init_func=init, blit=True)

    plt.show()

    # ani.save("md.gif")



"""
    Theory.ModelData.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ElutionCurveModels import emg
from .SolidSphere import a_func, b_to_a_func

class ModelData:
    def __init__(self, qv, num_el, Rg):
        self.qv = qv
        self.cx = np.arange(num_el)

        h = 1
        mu = num_el//2
        sigma = 30
        tau = 0
        self.e_params = [(h, mu, sigma, tau)]
        self.cy = emg(self.cx, h, mu, sigma, tau)
        self.Rg = Rg
        self.R = R = np.sqrt(5/3)*Rg

        self.a = a_func(qv, R)
        self.b2a = b_to_a_func(qv, R)
        self.b = self.a * self.b2a

    def get_data(self, noise=0.01, conc=0, seed=None):
        c = self.cy/np.max(self.cy)
        if conc > 0:
            P = np.array([self.a, self.b*conc]).T
            C = np.array([c, c**2])
        else:
            P = self.a[:, np.newaxis]
            C = c[np.newaxis,:]

        M = P @ C
        print('M.shape', M.shape)
        M_ = M/np.max(M)

        if seed is not None:
            np.random.seed(seed)

        if noise > 0:
            GE = np.random.normal(0, 1, M.shape)*noise/M_
            M *= (1 + GE)
            E = np.abs(M*GE)

        return M, E

def demo1():
    from MatrixData import simple_plot_3d

    fontsize = 16
    noise = 0.01

    qv = np.linspace(0.005, 0.5, 600)
    md = ModelData(qv, 300, 32)
    M, E = md.get_data(noise=noise, conc=1)

    fig = plt.figure(figsize=(21, 7))
    ax1 = fig.add_subplot(131)
    ax2 = fig.add_subplot(132)
    ax3 = fig.add_subplot(133, projection='3d')

    fig.suptitle("Noisy Data generated with Solid Sphere Model", fontsize=20)

    ax1.set_title("Gaussian Elution Curve", fontsize=fontsize)
    ax1.plot(md.cx, md.cy, color='orange')

    ax2.set_title("Solid Sphere Scattering Curve: Rg=%g (R=%.3g)" % (md.Rg, md.R), fontsize=fontsize)
    ax2.set_yscale('log')
    ax2.plot(qv, md.a, label='A(q)')

    ax2t = ax2.twinx()
    ax2t.grid(False)
    ax2t.set_ylim(-2, 2)
    ax2t.plot(qv, md.b2a, color='yellow', label='B(q)/A(q)')
    ax2t.plot(qv, md.b, color='pink', label='B(q)')

    ax2.legend(fontsize=fontsize)
    ax2t.legend(bbox_to_anchor=(1, 0.9), loc='upper right', fontsize=fontsize)

    ax3.set_title("Generated Noisy Data (noise=%g)" % noise, fontsize=fontsize)
    simple_plot_3d(ax3, M, x=qv)

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()

def demo2():
    from Rank.Synthesized import synthesized_data

    qv = np.linspace(0.005, 0.5, 100)
    qi = bisect_right(qv, 0.02)

    Rg = 32
    qv = np.linspace(0.005, 0.5, 600)
    md = ModelData(qv, 300, Rg)

    noises = [0.01, 0.03, 0.06]
    boundaries = [0.05, 0.1, 0.15, 0.2, 0.25]

    nrows = len(noises)
    ncols = 5
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(21, 11))

    fig.suptitle("Rank Boundary Evaluation with Noisy Solid Sphere Model", fontsize=20)

    for i, noise in enumerate(noises):
        for j, b in enumerate(boundaries):
            M, E = md.get_data(noise=noise, conc=1)
            c = M[qi,:]
            c = c/np.max(c)
            C = np.array([c, c**2])
            M_ = synthesized_data(qv, M, boundary=b)
            P = M_ @ np.linalg.pinv(C)
            a_ = P[:,0]
            b_ = P[:,1]
            ax = axes[i,j]
            ax.plot(qv, a_)
            axt = ax.twinx()
            axt.grid(False)
            axt.plot(qv, b_, color='pink')

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)

    plt.show()



"""
    EoiiPlotUtils.py.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt

def plot_eoii(qv, jv, D, P, C, axes):

    i = bisect_right(qv, 0.02)

    cy = C[0,:]
    by = C[1,:]
    mu = np.argmax(cy)
    py = D[:,mu]
    aq, bq = P.T

    P_ = D @ np.linalg.pinv(C)
    aq_, bq_ = P_.T

    ax1 = axes[0]
    ax2 = axes[1]
    ax3 = axes[2]

    ax1.plot(jv, D[i,:], label="apparent data")
    ax1.plot(jv, aq[i]*cy, label="concentration")
    ax1.plot(jv, bq[i]*by, color="pink", label="eoii")
    ax1.legend()

    ax3.set_yscale("log")
    for ax in ax2, ax3:

        ax.plot(qv, py, label="apparent data")
        ax.plot(qv, aq, label="true aq")
        ax.plot(qv, bq, color="pink", label="true bq")
        ax.plot(qv, aq_, ":", label="solved aq")
        ax.plot(qv, bq_, ":", label="solved bq")

        ymin, ymax = ax2.get_ylim()
        ax.set_ylim(ymin, ymax)

        q = qv[i]
        ax.plot([q,q], [ymin, ymax], color="yellow", label="q=0.02")

        ax.legend()



"""
    Models.Stochastic.DispersiveStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

def study(x, y, baseline, model, peaks, peak_rgs, qualities, props, curent_info=None, debug=False):
    from importlib import reload
    import Models.Stochastic.DispersiveMonopore
    reload(Models.Stochastic.DispersiveMonopore)
    from molass_legacy.Models.Stochastic.DispersiveMonopore import guess_params_using_moments
    import Models.Stochastic.DispersiveUtils
    reload(Models.Stochastic.DispersiveUtils)
    from molass_legacy.Models.Stochastic.DispersiveUtils import compute_elution_curves

    egh_moments_list = compute_egh_moments(peaks)
    ret = guess_params_using_moments(x, y, egh_moments_list, peak_rgs, qualities, props, debug=debug)
    if ret is None:
        return
    sdm_params, temp_rgs, bounds = ret

    show_edm_info = False
    if curent_info is not None:
        curent_model, modelname, params_array = curent_info
        if modelname == "EDM":
            show_edm_info = True

    in_folder = get_in_folder()
    is_oa_ald = in_folder.find('sample_data') >= 0

    with plt.Dp():
        if show_edm_info:
            fig, (ax1, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(20,4))
            fig.suptitle("Study on SDM-EDM Relation with " + in_folder, fontsize=20)
        else:
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))

        ax1.set_title("EGH Decomposition", fontsize=16)
        ax1.plot(x, y)
        cy_list = []
        for k, params in enumerate(peaks):
            cy = model(x, params)
            cy_list.append(cy)
            ax1.plot(x, cy, ":", label='component-%d' % k)
        ty = np.sum(cy_list, axis=0)
        ax1.plot(x, ty, ":", color="red", label='model total')
        ax1.legend()

        ax2.set_title("SDM Decomposition", fontsize=16)
        cy_list, ty = compute_elution_curves(x, sdm_params, temp_rgs)
        ax2.plot(x, y)
        for k, cy in enumerate(cy_list):
            ax2.plot(x, cy, ":", label='component-%d' % k)
        ax2.plot(x, ty, ":", color="red", label='model total')
        ax2.legend()

        if show_edm_info:
            ax3.set_title("EDM Decomposition", fontsize=16)
            ax3.plot(x, y)
            cy_list = []
            for k, params in enumerate(params_array):
                cy = curent_model(x - x[0], params)
                ax3.plot(x, cy, ":", label='component-%d' % k)
                cy_list.append(cy)
            ty = np.sum(cy_list, axis=0)
            ax3.plot(x, ty, ":", color="red", label='model total')
            ax3.legend()

            # N, K, x0, poresize, N0, tI
            poresize = sdm_params[3]
            # t0, u, a, b, e, Dz, cinj
            porosities = params_array[:,4]
            edm_rgs = poresize * (1 - np.power(porosities, 1/3))

            ind = np.arange(len(peak_rgs))  # the x locations for the groups
            width = 0.35  # the width of the bars

            ax4.set_title("$K_{SEC} \sim Porosity$ Proof with Poresize=%.3g" % poresize, fontsize=16)
            ax4.bar(ind - width/2, peak_rgs, width, label="SAXS $R_g$")
            ax4.bar(ind + width/2, edm_rgs, width, label="Kinetic $R_g$")

            ax4.set_ylabel("$R_g$")
            ax4.set_xticks(np.arange(len(peak_rgs)))
            if is_oa_ald:
                labels = ('OA', 'OA-r', 'ALD')
            else:
                labels = ['$P_%d$' % k for k in range(len(peak_rgs))]
            ax4.set_xticklabels(labels)
            ax4.set_ylim(0, max(60, np.max(peak_rgs)*1.1))
            ax4.legend(loc="upper left")
            axt = ax4.twinx()
            axt.grid(False)
            axt.set_ylabel("Porosity")
            axt.plot(ind, porosities, "o", color="red", label="Porosity")
            axt.set_ylim(0, 1)
            axt.legend()

        fig.tight_layout()
        plt.show()



"""
    Alsaker2017.py

    converted from file1.R in     
"""
import numpy as np
from scipy.stats import norm
from scipy.optimize import minimize
from scipy.signal import butter, filtfilt
from scipy.interpolate import UnivariateSpline
from scipy.linalg import toeplitz
from scipy.sparse import diags
from scipy.linalg import sqrtm
from sklearn.linear_model import LinearRegression
import matplotlib.pyplot as plt

def estimate_Rg(M, num_reps, starting_value=-1):
    if isinstance(starting_value, int):
        starting_value = [starting_value] * num_reps
    
    for val in starting_value:
        if val < -1:
            print("Please enter positive initial values.")
            return
    
    if starting_value[0] != -1:
        choose_sp = True
        print("Initial points input by user.")
        if num_reps > 1:
            for i, val in enumerate(starting_value, start=1):
                if val == 1:
                    print(f"Removed zero points from replicate {i}")
                elif val == 2:
                    print(f"Removed first point from replicate {i}")
                elif val > 2:
                    print(f"Removed first {val-1} points from replicate {i}")
    else:
        choose_sp = False
        starting_value = [1] * num_reps
    
    sp = starting_value
    nreps = num_reps
    M = np.asarray(M)
    
    if nreps == 1:
        # check for zero values in M[:,1]
        first = True
        if M[0, 1] == 0 and first:
            first = False
            i = 0
            while M[i, 1] == 0 and i < len(M[:, 1]) - 1:
                i += 1
            num = i
        print("num=", num)
        if not first:
            # temp_M = M[num+1:, :]
            temp_M = M[num:, :]     # modified
            len_ = len(M[:, 0]) - num
            M = np.zeros((len_, 3))
            M[:, 0] = temp_M[:, 0]
            M[:, 1] = temp_M[:, 1]
            M[:, 2] = temp_M[:, 2]
            n = len_
            out_e = f"Warning: first {num} intensity values equal zero. These values were stripped from the data."
            print(out_e)
        
        first = True
        for i, val in enumerate(M[:, 1]):
            if val < 0 and first:
                n = i
                first = False
                print("Warning: negative intensity values found.")
        if first:
            n = len(M[:, 1])
        
        y3 = np.diff(np.diff(np.diff(np.diff(np.log(M[sp[0]-1:n, 1])))))
        cp2i = cpt_var(y3, n, 7)
        
        if isinstance(cp2i, int) and cp2i == 0:
            cp2i = n - 1
        if cp2i < 60:
            cp2i = 60
        if cp2i > 120:
            cp2i = 120
        
        cp2 = cp2i + sp[0] - 1
        m = 0
        
        comb_spline_fit = comb_spline(M, cp2, cp2i, nreps, sp, m)
        gamma = np.zeros(cp2)
        out = ind_ar_struc(comb_spline_fit, M[sp[0]-1:cp2+sp[0], 1], cp2)
        gamma = out[1:]
        arsum = out[0]
        sigma = create_gamma_matrix(gamma, cp2)
        t = b_v_tradeoff_comb(M, cp2, sigma, comb_spline_fit, nreps, sp, arsum)
        output = calc_Rg(M[:n, :], sigma, t, cp2, nreps, sp, arsum, comb_spline_fit, choose_sp)
        return output
    
    if nreps >= 2:
        n = len(M[:, 1])
        if np.any(M[:, 1] <= 0):
            n = np.where(M[:, 1] <= 0)[0][0]
        for i in range(2, nreps+1):
            if np.any(M[:, i] <= 0):
                temp = np.where(M[:, i] <= 0)[0][0]
                if temp < n:
                    n = temp
        
        cp2i = np.zeros(nreps, dtype=int)
        
        for i in range(nreps):
            y3 = np.diff(np.diff(np.diff(np.diff(np.log(M[sp[i]-1:n, i+1])))))
            cp2i[i] = cpt_var(y3, n, 7)
            if cp2i[i] < 60:
                cp2i[i] = 60
            if cp2i[i] > 120:
                cp2i[i] = 120
        
        cp2 = cp2i + sp - 1
        
        n = min(cp2) - max(sp) + 2
        m = max(cp2) - min(sp) + 1
        
        comb_spline_fit = comb_spline(M, cp2, cp2i, nreps, sp, m)
        gamma = np.zeros((m, nreps))
        arsum = 0
        for i in range(nreps):
            out = ind_ar_struc(comb_spline_fit[(sp[i]-min(sp)):(cp2[i]-min(sp)+1), i], M[sp[i]-1:cp2[i], i+1], m)
            gamma[:, i] = out[1:]
            arsum += out[0]
        arsum /= nreps
        avg_gamma = np.mean(gamma, axis=1)
        
        sigma = create_gamma_matrix(avg_gamma, m)
        t = b_v_tradeoff_comb(M, n, sigma, comb_spline_fit, nreps, sp, arsum)
        cp2 = t + sp - 1
        ep = cp2 + 1
        
        output = calc_Rg(M, sigma, t, n, nreps, sp, arsum, comb_spline_fit, choose_sp)
        
        return output

def comb_spline(M, cp2, cp2i, nreps, sp, m):
    import numpy as np
    from scipy.interpolate import UnivariateSpline
    
    df = 6
    if m > 80:
        df = 8
    
    if nreps > 2:
        ncp = UnivariateSpline(M[min(sp)-1:max(cp2), 0], np.arange(min(sp), max(cp2)+1), k=df)
        
        X = np.ones((cp2[0]-sp[0]+1, 1))
        X = np.hstack((X, np.zeros((cp2[0]-sp[0]+1, nreps-1))))
        X = np.hstack((X, ncp(M[sp[0]-1:cp2[0], 0])))
        Y = np.log(M[sp[0]-1:cp2[0], 1])
        for i in range(1, nreps-1):
            Y = np.concatenate((Y, np.log(M[sp[i]-1:cp2[i], i+1])))
            temp = np.zeros((cp2[i]-sp[i]+1, i))
            temp = np.hstack((temp, np.ones((cp2[i]-sp[i]+1, 1))))
            temp = np.hstack((temp, np.zeros((cp2[i]-sp[i]+1, nreps-i-1))))
            temp = np.hstack((temp, ncp(M[sp[i]-1:cp2[i], 0])))
            X = np.vstack((X, temp))
        Y = np.concatenate((Y, np.log(M[sp[nreps-1]-1:cp2[nreps-1], nreps])))
        temp = np.zeros((cp2[nreps-1]-sp[nreps-1]+1, nreps-1))
        temp = np.hstack((temp, np.ones((cp2[nreps-1]-sp[nreps-1]+1, 1))))
        temp = np.hstack((temp, ncp(M[sp[nreps-1]-1:cp2[nreps-1], 0])))
        X = np.vstack((X, temp))
        beta_est = np.linalg.lstsq(X, Y, rcond=None)[0]
        X = np.ones(m)
        for j in range(1, nreps):
            X = np.hstack((X, np.zeros(m)))
        X = np.hstack((X, ncp(np.linspace(M[min(sp)-1, 0], M[max(cp2)-1, 0], m))))
        beta_curve = X @ beta_est
        
        for i in range(1, nreps):
            X = np.zeros(m)
            for j in range(1, nreps):
                if j == i:
                    X = np.hstack((X, np.ones(m)))
                else:
                    X = np.hstack((X, np.zeros(m)))
            X = np.hstack((X, ncp(np.linspace(M[min(sp)-1, 0], M[max(cp2)-1, 0], m))))
            beta_curve_temp = X @ beta_est
            beta_curve = np.vstack((beta_curve, beta_curve_temp))
        return beta_curve.T
    
    if nreps == 2:
        ncp = UnivariateSpline(M[min(sp)-1:max(cp2), 0], np.arange(min(sp), max(cp2)+1), k=df)
        
        X = np.ones((cp2[0]-sp[0]+1, 1))
        X = np.hstack((X, np.zeros((cp2[0]-sp[0]+1, nreps-1))))
        X = np.hstack((X, ncp(M[sp[0]-1:cp2[0], 0])))
        Y = np.log(M[sp[0]-1:cp2[0], 1])
        
        Y = np.concatenate((Y, np.log(M[sp[1]-1:cp2[1], 2])))
        temp = np.zeros((cp2[1]-sp[1]+1, nreps-1))
        temp = np.hstack((temp, np.ones((cp2[1]-sp[1]+1, 1))))
        temp = np.hstack((temp, ncp(M[sp[1]-1:cp2[1], 0])))
        X = np.vstack((X, temp))
        
        beta_est = np.linalg.lstsq(X, Y, rcond=None)[0]
        X = np.ones(m)
        X = np.hstack((X, np.zeros(m)))
        X = np.hstack((X, ncp(np.linspace(M[min(sp)-1, 0], M[max(cp2)-1, 0], m))))
        beta_curve = X @ beta_est
        
        X = np.zeros(m)
        X = np.hstack((X, np.ones(m)))
        X = np.hstack((X, ncp(np.linspace(M[min(sp)-1, 0], M[max(cp2)-1, 0], m))))
        beta_curve_temp = X @ beta_est
        
        beta_curve = np.vstack((beta_curve, beta_curve_temp))
        
        return beta_curve.T
    
    if nreps == 1:
        X = np.ones((cp2-sp[0]+1, 1))
        # X = np.hstack((X, UnivariateSpline(M[sp[0]-1:cp2+sp[0], 0], np.arange(sp[0], cp2+sp[0]+1), k=df)(M[sp[0]-1:cp2+sp[0], 0])))
        X = np.hstack((X, UnivariateSpline(M[sp[0]-1:cp2+sp[0], 0], np.arange(sp[0], cp2+sp[0]+1))(M[sp[0]-1:cp2+sp[0], 0])))
        Y = np.log(M[sp[0]-1:cp2+sp[0], 1])
        
        beta_est = np.linalg.lstsq(X, Y, rcond=None)[0]
        beta_curve = X @ beta_est
        
        return beta_curve

def ind_ar_struc(comb_spline_fit, d2, m):
    import numpy as np
    from statsmodels.tsa.ar_model import AutoReg
    
    resid = np.log(d2) - comb_spline_fit
    ar_0 = AutoReg(resid, lags=5, old_names=False).fit()
    if ar_0.k_ar == 0:
        phi_temp = 0
    else:
        phi_temp = ar_0.ar_params
    a_0 = {'phi': phi_temp, 'theta': np.array([0]), 'sigma2': ar_0.sigma2}
    arsum = ar_0.sigma2 / (1 - np.sum(ar_0.ar_params))**2
    
    g_0 = np.concatenate(([arsum], ar_0.acf(len=m)[1:]))
    return g_0

def create_gamma_matrix(g, p):
    import numpy as np
    
    gamma = np.zeros((p, p))
    for i in range(p):
        for j in range(p):
            gamma[i, j] = g[abs(i-j)]
    return gamma

def cpt_var(y3, n, pen_value):
    import ruptures as rpt
    import molass_legacy.KekLib.DebugPlot as dpl
    
    with dpl.Dp():
        fig, ax = dpl.subplots()
        ax.set_title("cpt_var")
        ax.plot(y3)
        fig.tight_layout()
        dpl.show()

    # cost = numpy_norm_outer(y3)
    # n_cp = pelt(cost, jump=pen_value*np.log(n), min_size=60, n_samples=len(y3))
    # return n_cp.sum()
    # my_bkps = rpt.Pelt(model="l1", jump=int(pen_value*np.log(n)), min_size=60).fit(y3).predict(pen_value*np.log(n))
    # my_bkps = rpt.Pelt(model="l1", jump=5, min_size=3).fit(y3).predict(pen_value*np.log(n))
    my_bkps = rpt.Dynp(model="normal", params={"add_small_diag": False}).fit(y3).predict(pen_value)
    print("my_bkps=", my_bkps, "len(my_bkps)=", len(my_bkps))
    return my_bkps[0]

def b_v_tradeoff_comb(M, cp2, sigma, comb_spline_fit, nreps, sp, arsum):
    #initialize variables
    f = np.full(cp2, 10000000)
    var_est = np.full(cp2, 10000000)
    sum_bias2_avg = np.full(cp2, 10000000)

    s = M[:,0]
    delta_s = np.mean(np.diff(s[:19]))

    #loop though data up to cp2
    for k in range(2, cp2):
        if nreps > 2:
            tempk = k
            k = max(sp) + k - 1
            len_ = np.repeat(k, nreps) - np.repeat(min(sp), nreps) + 2
            lenb = len_ - (sp-min(sp)+1) + 1

            #fit quadratic curve over data
            X = np.column_stack((np.repeat(1, lenb[0]), np.zeros((lenb[0], nreps-1)), M[sp[0]:(k+1),0]**2, M[sp[0]:(k+1),0]**4))
            Y = np.log(M[sp[0]:(k+1),1])

            for i in range(1, nreps-1):
                Y = np.concatenate((Y, np.log(M[sp[i]:(k+1),i+1])))
                temp = np.column_stack((np.zeros((lenb[i], i-1)), np.repeat(1, lenb[i]), np.zeros((lenb[i], nreps-i)), M[sp[i]:(k+1),0]**2, M[sp[i]:(k+1),0]**4))
                X = np.vstack((X, temp))

            Y = np.concatenate((Y, np.log(M[sp[nreps]:(k+1),nreps+1])))
            temp = np.column_stack((np.zeros((lenb[nreps], nreps-1)), np.repeat(1, lenb[nreps]), M[sp[nreps]:(k+1),0]**2, M[sp[nreps]:(k+1),0]**4))
            X = np.vstack((X, temp))

        if nreps == 2:
            tempk = k
            k = max(sp) + k - 1
            len_ = np.repeat(k, nreps) - np.repeat(min(sp), nreps) + 2
            lenb = len_ - (sp-min(sp)+1) + 1

            #fit quadratic curve over data
            X = np.column_stack((np.repeat(1, lenb[0]), np.zeros((lenb[0], nreps-1)), M[sp[0]:(k+1),0]**2, M[sp[0]:(k+1),0]**4))
            Y = np.log(M[sp[0]:(k+1),1])

            Y = np.concatenate((Y, np.log(M[sp[nreps]:(k+1),nreps+1])))
            temp = np.column_stack((np.zeros((lenb[nreps], nreps-1)), np.repeat(1, lenb[nreps]), M[sp[nreps]:(k+1),0]**2, M[sp[nreps]:(k+1),0]**4))
            X = np.vstack((X, temp))

        if nreps == 1:
            tempk = k
            #fit quadratic curve over data
            X = np.column_stack((np.repeat(1, k), M[0:k,0]**2, M[0:k,0]**4))
            Y = np.log(M[0:k,1])

        fit = LinearRegression(fit_intercept=False).fit(X, Y)
        alpha = fit.coef_

        #calculate the estimated bias of Rg
        sum_bias2_avg[tempk] = 9/784*(24*alpha[nreps+2])**2*(k*delta_s)**4

        #calculate variance of Rg
        var_est[tempk] =  (405*arsum)/(nreps*4*k**5*delta_s**4)   

        #bias-variance criterion
        f[tempk] = var_est[tempk] + sum_bias2_avg[tempk]

    #calculate minimum of bias-variance criterion
    t = np.where(f == np.min(f))[0][0]
    return t

def calc_Rg(M, sigma, t, cp2, nreps, sp, arsum, comb_spline_fit, choose_sp):
    s = M[:,0]
    delta_s = np.mean(np.diff(s[:10]))
    output = np.zeros(2)

    if nreps > 2:
        t = max(sp) + t - 1
        len_ = np.repeat(t, nreps) - np.repeat(min(sp), nreps) + 2
        len2 = len_ - (sp-min(sp)+1) + 1
        len3 = len_[0] + min(sp) - 1

        #create design matrix
        X = np.column_stack((np.repeat(1, len2[0]), np.zeros((len2[0], nreps-1)), M[sp[0]:(t+1),0]**2))
        Y = np.log(M[sp[0]:(t+1),1])

        for i in range(1, nreps-1):
            Y = np.concatenate((Y, np.log(M[sp[i]:(t+1),i+1])))
            temp = np.column_stack((np.zeros((len2[i], i-1)), np.repeat(1, len2[i]), np.zeros((len2[i], nreps-i)), M[sp[i]:(t+1),0]**2))
            X = np.vstack((X, temp))

        Y = np.concatenate((Y, np.log(M[sp[nreps]:(t+1),nreps+1])))
        temp = np.column_stack((np.zeros((len2[nreps], nreps-1)), np.repeat(1, len2[nreps]), M[sp[nreps]:(t+1),0]**2))
        X = np.vstack((X, temp))

        #Create covariance matrix
        get = np.column_stack((sigma[(sp[0]-min(sp)+1):len_[0],(sp[0]-min(sp)+1):len_[0]], np.zeros((sum(len2[1:]), len2[0]))))
        for i in range(1, nreps-1):
            temp = np.column_stack((np.zeros((sum(len2[1:i]), len2[i])), sigma[(sp[i]-min(sp)+1):len_[i],(sp[i]-min(sp)+1):len_[i]], np.zeros((sum(len2[(i+1):nreps]), len2[i]))))
            get = np.vstack((get, temp))
        temp = np.column_stack((np.zeros((sum(len2[1:(nreps-1)])), len2[nreps])), sigma[(sp[nreps]-min(sp)+1):len_[nreps],(sp[nreps]-min(sp)+1):len_[nreps]])
        get = np.vstack((get, temp))

        gamma_est = get

    if nreps == 2:
        t = max(sp) + t - 1
        len_ = np.repeat(t, nreps) - np.repeat(min(sp), nreps) + 2
        len2 = len_ - (sp-min(sp)+1) + 1
        len3 = len_[0] + min(sp) - 1

        #create design matrix
        X = np.column_stack((np.repeat(1, len2[0]), np.zeros((len2[0], nreps-1)), M[sp[0]:(t+1),0]**2))
        Y = np.log(M[sp[0]:(t+1),1])

        Y = np.concatenate((Y, np.log(M[sp[nreps]:(t+1),nreps+1])))
        temp = np.column_stack((np.zeros((len2[nreps], nreps-1)), np.repeat(1, len2[nreps]), M[sp[nreps]:(t+1),0]**2))
        X = np.vstack((X, temp))

        #create covariance matrix
        get = np.column_stack((sigma[(sp[0]-min(sp)+1):len_[0],(sp[0]-min(sp)+1):len_[0]], np.zeros((sum(len2[1:]), len2[0]))))
        temp = np.column_stack((np.zeros((sum(len2[1]), len2[nreps])), sigma[(sp[nreps]-min(sp)+1):len_[nreps],(sp[nreps]-min(sp)+1):len_[nreps]]))
        get = np.vstack((get, temp))

        gamma_est = get

    if nreps == 1:
        len3 = t
        X = np.column_stack((np.repeat(1, t), M[sp[0]:(t+sp[0]-1),0]**2))
        Y = np.log(M[sp[0]:(sp[0]+t-1),1])
        gamma_est = sigma[:t,:t]

    fit = LinearRegression(fit_intercept=False).fit(X, Y)
    alpha_curve = fit.predict(X)
    alpha_curve2 = fit.coef_[0] + fit.coef_[nreps]*M[min(sp):len3,0]**2

    e, V = np.linalg.eig(gamma_est)
    B = V @ sqrtm(np.diag(e)) @ V.T
    Y = B @ Y
    X = B @ X
    fit = LinearRegression(fit_intercept=False).fit(X, Y)

    if nreps == 1:
        resids = np.log(M[:cp2,1]) - fit.coef_[0] - fit.coef_[1]*M[:cp2,0]**2
    if nreps > 1:
        resids = np.zeros((cp2, nreps))
        for i in range(nreps):
            resids[:,i] = np.log(M[:cp2,i+1]) - fit.coef_[i] - fit.coef_[nreps]*M[:cp2,0]**2

    #estimate Rg using standard regression technique
    alpha = fit.coef_
    #check for negative Rg
    if alpha[nreps] > 0:
        print("Negative Rg value found. Program stopped.")

    Rg_prev = (-3*alpha[nreps])

    #############################
    #Outlier Detection
    #############################
    if choose_sp == False:
        NUM = t//2
        DFBETAS = np.zeros(NUM)
        d_i = np.zeros(NUM)
        s_d_i = np.zeros(NUM)
        MSE = np.sum((Y-fit.predict(X))**2)/(t-3)
        Y_hat = fit.predict(X)
        d2 = np.log(M[:,1])
        d1 = M[:,0]
        bool_ = True
        i = 0

        while bool_ == True and i < NUM:
            d2_i = d2[1:i+1]
            d1_i = d1[1:i+1]

            M_i = M[i+1:,:]
            comb_spline_fit_i = comb_spline_fit[i+1:]

            t_i = b_v_tradeoff_comb(M_i, cp2-i, sigma, comb_spline_fit_i, nreps, sp, arsum)

            Y_i = np.log(d2_i[:t_i])
            X_i = np.column_stack((np.repeat(1, t_i), (d1_i[:t_i])**2))

            if nreps > 2:
                len_ = np.repeat(t_i, nreps) - np.repeat(1, nreps) + 2
                len2 = len_

                #create design matrix
                X_i = np.column_stack((np.repeat(1, len2[0]), np.zeros((len2[0], nreps-1)), M_i[:t_i+1,0]**2))
                Y_i = np.log(M_i[:t_i+1,1])

                for i in range(1, nreps-1):
                    Y_i = np.concatenate((Y_i, np.log(M[:t_i+1,i+1])))
                    temp = np.column_stack((np.zeros((len2[i], i-1)), np.repeat(1, len2[i]), np.zeros((len2[i], nreps-i)), M_i[:t_i+1,0]**2))
                    X_i = np.vstack((X_i, temp))

                Y_i = np.concatenate((Y_i, np.log(M_i[:t_i+1,nreps+1])))
                temp = np.column_stack((np.zeros((len2[nreps], nreps-1)), np.repeat(1, len2[nreps]), M_i[:t_i+1,0]**2))
                X_i = np.vstack((X_i, temp))

            fit_i = LinearRegression(fit_intercept=False).fit(X_i, Y_i)
            beta_i = fit_i.coef_[nreps]

            Rg_now = (-3*beta_i)
            var_Rg2 = np.sqrt((405*arsum)/(nreps*4*t_i**5*delta_s**4)/9)

            DFBETAS[i] = np.abs(Rg_prev - Rg_now)/np.sqrt(np.var(Y_i))
            Rg_prev = Rg_now
            if DFBETAS[i] < 2/np.sqrt(t_i):
                bool_ = False
            if DFBETAS[i] >= 2/np.sqrt(t_i):
                i += 1

        if i > 0:
            print("Initial points selected automatically.")
            if nreps == 1:
                if i == 1:
                    print("Removed first point from the curve.")
                else:
                    print(f"Removed first {i-1} points from the curve.")
            if nreps > 1:
                if i == 1:
                    print("Removed first point from each curve.")
                else:
                    print(f"Removed first {i-1} points from each curve.")
            sp = np.repeat(i, nreps)
            s = M[:,0]
            delta_s = np.mean(np.diff(s[:10]))
            output = np.zeros(2)

            if nreps > 2:
                t = max(sp) + t - 1
                len_ = np.repeat(t, nreps) - np.repeat(min(sp), nreps) + 2
                len2 = len_ - (sp-min(sp)+1) + 1
                len3 = len_[0] + min(sp) - 1

                #create design matrix
                X = np.column_stack((np.repeat(1, len2[0]), np.zeros((len2[0], nreps-1)), M[sp[0]:(t+1),0]**2))
                Y = np.log(M[sp[0]:(t+1),1])

                for i in range(1, nreps-1):
                    Y = np.concatenate((Y, np.log(M[sp[i]:(t+1),i+1])))
                    temp = np.column_stack((np.zeros((len2[i], i-1)), np.repeat(1, len2[i]), np.zeros((len2[i], nreps-i)), M[sp[i]:(t+1),0]**2))
                    X = np.vstack((X, temp))

                Y = np.concatenate((Y, np.log(M[sp[nreps]:(t+1),nreps+1])))
                temp = np.column_stack((np.zeros((len2[nreps], nreps-1)), np.repeat(1, len2[nreps]), M[sp[nreps]:(t+1),0]**2))
                X = np.vstack((X, temp))

                #Create covariance matrix
                get = np.column_stack((sigma[(sp[0]-min(sp)+1):len_[0],(sp[0]-min(sp)+1):len_[0]], np.zeros((sum(len2[1:]), len2[0]))))
                for i in range(1, nreps-1):
                    temp = np.column_stack((np.zeros((sum(len2[1:i]), len2[i])), sigma[(sp[i]-min(sp)+1):len_[i],(sp[i]-min(sp)+1):len_[i]], np.zeros((sum(len2[(i+1):nreps]), len2[i]))))
                    get = np.vstack((get, temp))
                temp = np.column_stack((np.zeros((sum(len2[1:(nreps-1)])), len2[nreps])), sigma[(sp[nreps]-min(sp)+1):len_[nreps],(sp[nreps]-min(sp)+1):len_[nreps]])
                get = np.vstack((get, temp))

                gamma_est = get

            if nreps == 2:
                t = max(sp) + t - 1
                len_ = np.repeat(t, nreps) - np.repeat(min(sp), nreps) + 2
                len2 = len_ - (sp-min(sp)+1) + 1
                len3 = len_[0] + min(sp) - 1

                #create design matrix
                X = np.column_stack((np.repeat(1, len2[0]), np.zeros((len2[0], nreps-1)), M[sp[0]:(t+1),0]**2))
                Y = np.log(M[sp[0]:(t+1),1])

                Y = np.concatenate((Y, np.log(M[sp[nreps]:(t+1),nreps+1])))
                temp = np.column_stack((np.zeros((len2[nreps], nreps-1)), np.repeat(1, len2[nreps]), M[sp[nreps]:(t+1),0]**2))
                X = np.vstack((X, temp))

                #create covariance matrix
                get = np.column_stack((sigma[(sp[0]-min(sp)+1):len_[0],(sp[0]-min(sp)+1):len_[0]], np.zeros((sum(len2[1:]), len2[0]))))
                temp = np.column_stack((np.zeros((sum(len2[1]), len2[nreps])), sigma[(sp[nreps]-min(sp)+1):len_[nreps],(sp[nreps]-min(sp)+1):len_[nreps]]))
                get = np.vstack((get, temp))

                gamma_est = get

            if nreps == 1:
                len3 = t
                X = np.column_stack((np.repeat(1, t), M[sp[0]:(t+sp[0]-1),0]**2))
                Y = np.log(M[sp[0]:(sp[0]+t-1),1])
                gamma_est = sigma[:t,:t]

            fit = LinearRegression(fit_intercept=False).fit(X, Y)
            alpha_curve = fit.predict(X)
            alpha_curve2 = fit.coef_[0] + fit.coef_[nreps]*M[min(sp):len3,0]**2

            e, V = np.linalg.eig(gamma_est)
            B = V @ sqrtm(np.diag(e)) @ V.T
            Y = B @ Y
            X = B @ X
            fit = LinearRegression(fit_intercept=False).fit(X, Y)

            if nreps == 1:
                resids = np.log(M[:cp2,1]) - fit.coef_[0] - fit.coef_[1]*M[:cp2,0]**2
            if nreps > 1:
                resids = np.zeros((cp2, nreps))
                for i in range(nreps):
                    resids[:,i] = np.log(M[:cp2,i+1]) - fit.coef_[i] - fit.coef_[nreps]*M[:cp2,0]**2

            #estimate Rg using standard regression technique
            alpha = fit.coef_
            #check for negative Rg
            if alpha[nreps] > 0:
                print("Negative Rg value found. Program stopped.")

    ############################

    output[0] = np.sqrt(-3*alpha[nreps])

    #approximate the variance of Rg hat using Taylor linearization
    var_Rg2 = (405*arsum)/(nreps*4*t**5*delta_s**4)/9   
    output[1] = np.sqrt(-3/(4*alpha[nreps])*var_Rg2)

    if nreps == 1:
        #construct plots
        d1 = M[:,0]
        d2 = M[:,1]
        n = len(d1)

        plt.plot(d1[:cp2], resids, 'o', color='blue')
        plt.axhline(0, color='black')
        plt.plot(d1[(t+sp[0]):cp2], resids[(t+sp[0]):cp2], 'o', color='blue')
        plt.plot(d1[sp[0]:(sp[0]+t-1)], resids[sp[0]:(sp[0]+t-1)], 'o', color='red')
        if sp[0] > 1:
            plt.plot(d1[1:(sp[0]-1)], resids[1:(sp[0]-1)], 'o', color='blue')
        plt.legend(["Data points used to fit curve", "Excluded data points"])
        plt.xlabel("S")
        plt.ylabel("Residuals")
        plt.show()

        plt.plot(d1[(sp[0]+t):(sp[0]+cp2-1)]**2, np.log(d2[(sp[0]+t):(sp[0]+cp2-1)]), 'o', color='blue')
        plt.plot(d1[sp[0]:(sp[0]+t-1)]**2, np.log(d2[sp[0]:(sp[0]+t-1)]), 'o', color='red')
        plt.plot(d1[sp[0]:(sp[0]+t-1)]**2, alpha_curve, linewidth=2)
        plt.legend(["Data points used to fit curve", "Fitted curve"])
        plt.xlabel("S^2")
        plt.ylabel("Log(Intensity)")
        plt.show()

        plt.plot(d1[sp[0]:(sp[0]+cp2-1)], np.log(d2[sp[0]:(sp[0]+cp2-1)]), 'o', color='blue')
        plt.plot(d1[sp[0]:(sp[0]+t-1)], np.log(d2[sp[0]:(sp[0]+t-1)]), 'o', color='red')
        plt.plot(d1[sp[0]:(sp[0]+t-1)], alpha_curve, linewidth=2)
        plt.legend(["Data points used to fit curve", "Fitted curve"])
        plt.xlabel("S")
        plt.ylabel("Log(Intensity)")
        plt.show()

    if nreps > 1:
        #construct plots
        d1 = M[:,0]
        d2 = M[:,1]

        plt.plot(d1[:cp2], resids[:,0], 'o', color='blue')
        plt.axhline(0, color='black')
        for i in range(1, nreps):
            d1 = M[:,0]
            d2 = M[:,i+1]-fit.coef_[i]+fit.coef_[0]

            plt.plot(d1[(len3+1):cp2], resids[(len3+1):cp2,i], 'o', color='blue')
            plt.plot(d1[sp[i]:(len3)], resids[sp[i]:(len3),i], 'o', color='red')
            if sp[i] > 1:
                plt.plot(d1[1:(sp[i]-1)], resids[1:(sp[i]-1),i], 'o', color='blue')
        plt.legend(["Data points used to fit curve", "Excluded data points"])
        plt.xlabel("S")
        plt.ylabel("Residuals")
        plt.show()

        plt.plot(d1[(len3+1):(sp[0]+cp2-1)]**2, np.log(d2[(len3+1):(sp[0]+cp2-1)]), 'o', color='blue')
        plt.plot(d1[sp[0]:(len3)]**2, np.log(d2[sp[0]:(len3)]), 'o', color='red')
        plt.plot(d1[min(sp):(len3)]**2, alpha_curve2, linewidth=2)
        plt.legend(["Data points used to fit curve", "Fitted curve"])
        plt.xlabel("S^2")
        plt.ylabel("Log(Intensity)")
        plt.show()

        for i in range(1, nreps):
            d1 = M[:,0]
            d2 = M[:,i+1]-fit.coef_[i]+fit.coef_[0]

            plt.plot(d1[(len3+1):(sp[i]+cp2-1)]**2, np.log(d2[(len3+1):(sp[i]+cp2-1)]), 'o', color='blue')
            plt.plot(d1[sp[i]:(len3)]**2, np.log(d2[sp[i]:(len3)]), 'o', color='red')
            plt.plot(d1[min(sp):(len3)]**2, alpha_curve2, linewidth=2)
            plt.legend(["Data points used to fit curve", "Fitted curve"])
            plt.xlabel("S^2")
            plt.ylabel("Log(Intensity)")
            plt.show()

            plt.plot(d1[(len3+1):(sp[i]+cp2-1)], np.log(d2[(len3+1):(sp[i]+cp2-1)]), 'o', color='blue')
            plt.plot(d1[sp[i]:(len3)], np.log(d2[sp[i]:(len3)]), 'o', color='red')
            plt.plot(d1[min(sp):(len3)], alpha_curve2, linewidth=2)
            plt.legend(["Data points used to fit curve", "Fitted curve"])
            plt.xlabel("S")
            plt.ylabel("Log(Intensity)")
            plt.show()

    #return Rg and its standard deviation
    return np.array([output[0], output[1], t, cp2])


"""
    DefaultFont.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

from openpyxl.styles import DEFAULT_FONT
from molass_legacy._MOLASS.SerialSettings import get_setting

def set_default_font(name=None):
    if name is None:
        name = get_setting("report_default_font")

    if name is not None:
        DEFAULT_FONT.name = name

def set_default_bookfont(wb, name=None):
    # this is currently not used

    if name is None:
        name = get_setting("report_default_font")

    if name is not None:
        wb._named_styles['Normal'].font.name = name

def set_default_bookfont_com(wb, name=None):
    if name is None:
        name = get_setting("report_default_font")

    if name is not None:
        wb.Styles("Normal").Font.Name = name



"""

    GuiReferences.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF

"""
import os
from molass_legacy.KekLib.OurTkinter     import Tk
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy._MOLASS.Version import get_version_string

def get_doc_folder_url():
    from molass_legacy.KekLib.BasicUtils import get_home_folder
    # home = os.getcwd().replace( '\\', '/' )
    home = get_home_folder()
    url = 'file://' + home + '/doc'
    # print('doc_folder_url=', url)
    return url

browser = None

def get_default_browser():
    global browser
    if browser is None:
        import webbrowser
        browser = webbrowser.get()

    return browser

def get_preferable_browser():
    global browser
    if browser is None:
        import webbrowser
        try:
            # https://stackoverflow.com/questions/48056052/webbrowser-get-could-not-locate-runnable-browser
            chrome_path="C:\\Program Files (x86)\\Google\\Chrome\\Application\\chrome.exe"
            webbrowser.register('chrome', None, webbrowser.BackgroundBrowser(chrome_path))
            browser = webbrowser.get('chrome')
        except:
            browser = webbrowser.get()

    return browser

class GuiReferencesMenu(Tk.Menu):
    def __init__(self, parent, menubar ):
        self.parent = parent
        doc_folder_url = get_doc_folder_url()
        browser = get_preferable_browser()

        web_references = []

        if get_version_string().find("_MOLASS 2") >= 0:
            web_references += [("_MOLASS V2 Manipulation Guide", doc_folder_url + '/_MOLASS-Manipulation-Guide-2_0-draft.pdf')]

        web_references += [
            ("_MOLASS V1 User's Guide", doc_folder_url + '/_MOLASS-UsersGuide-1_0_13.pdf'),
            ("Formula (2.13), Feigin and Svergun 1987", doc_folder_url + '/InterferenceEffects-feigin-svergun-1987-excerpt.pdf'),
            # ('Automation of Guinier Analysis in SAXS', doc_folder_url + '/AutomationOfGuinierAnalysis.pdf'),
            # ('Bi-component Guinier-Porod Model', doc_folder_url + '/BicomponentGuinierPorodModel.pdf'),
            # ('A new Guinier–Porod model', 'https://www.ncnr.nist.gov/staff/hammouda/publications/2010_hammouda_j_appl_crys.pdf'),
            # ('Gaussian Processes', 'http://scikit-learn.org/stable/modules/gaussian_process.html'),
            ('Linear least squares (mathematics)', 'https://en.wikipedia.org/wiki/Linear_least_squares_%28mathematics%29#Weighted_linear_least_squares'),
            ('ATSAS AUTORG manual', 'http://www.embl-hamburg.de/biosaxs/manuals/autorg.html'),
            ('ATSAS ALMERGE manual', 'https://www.embl-hamburg.de/biosaxs/manuals/almerge.html'),
            ('ATSAS DATGNOM4 manual', 'https://www.embl-hamburg.de/biosaxs/manuals/datgnom.html'),
            ]

        Tk.Menu.__init__(self, menubar, tearoff=0 )
        menubar.add_cascade( label="References", menu=self )
        for label_, url_ in web_references:
            self.add_command( label=label_, command=lambda url=url_: browser.open_new( url ) )



"""
    Models.Stochastic.DispersivePdf.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.SecTheory.SecCF import dispersive_monopore
from molass_legacy.SecTheory.SecPDF import FftInvPdf
dispersive_monopore_pdf_impl = FftInvPdf(dispersive_monopore)

DEFUALT_TIMESCALE = 0.25    # 0.1 for FER_OA
N0 = 14400.0    # 48000*0.3 (30cm) or (t0/σ0)**2, see meeting document 20221104/index.html 

def dispersive_monopore_pdf(x, npi, tpi, N0, t0, timescale=DEFUALT_TIMESCALE):
    return timescale*dispersive_monopore_pdf_impl(timescale*x, npi, timescale*tpi, N0, timescale*t0)

def timescale_proof():
    import molass_legacy.KekLib.DebugPlot as plt
    N = 2000
    T = 0.5
    K = N*T
    me = 1.5
    mp = 1.5
    poresize = 75
    rgs = np.array([50, 40, 30, 20])
    rhov = rgs/poresize
    rhov[rhov > 1] = 1
    t0 = 500
    N0 = 14400
    t = np.arange(0, 1000)

    def compute_curves(x, N, T, x0, tI, N0, rhov, timescale=1):
        cy_list = []
        for rho in rhov:
            ni = N*(1 - rho)**me
            ti = T*(1 - rho)**mp
            cy = dispersive_monopore_pdf(x - tI, ni, ti, N0, x0 - tI, timescale=timescale)
            cy_list.append(cy)
        return cy_list

    rv = np.linspace(1, 70, 100)
    def compute_trv(N, T, poresize):
        rhov_ = rv/poresize
        rhov_[rhov_ > 1] = 1
        niv = N*(1 - rhov_)**me
        tiv = T*(1 - rhov_)**mp
        trv = t0 + niv*tiv
        return trv
    trv = compute_trv(N, T, poresize)

    def plot_curves(ax, x, x0, tI, title, with_arrow=True, timescale=1):
        ax.set_title(title + " with $t_{I}$=%g" % tI, fontsize=16)
        cy_list = compute_curves(x, N, T, x0, tI, N0, rhov, timescale=timescale)
        curves = []
        for k, (cy, rg) in enumerate(zip(cy_list, rgs)):
            curve, = ax.plot(x, cy, ':', label='$R_g$=%g' % rg)
            curves.append(curve)
        lines = []
        line = ax.axvline(x=x0, color='red', label='$t_0$')
        lines.append(line)
        line = ax.axvline(x=tI, color='gray', label='$t_I$')
        lines.append(line)
        if with_arrow:
            hy = np.average(ax.get_ylim())
            dx = x0 - tI
            ax.arrow(x=tI, y=hy, dx=dx, dy=0, width=0.005, head_width=0.015,
                    head_length=0.2*dx, length_includes_head=True, color='pink')
        ax.legend()

        axt = ax.twinx()
        axt.grid(False)
        excl_curve, = axt.plot(trv + tI, rv, color='yellow', label='exclusion curve')
        axt.legend(loc="center right")

        return curves, lines, excl_curve

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("SDM PDF Study in Experimental Time Axis with Poresize=%g, N*T=%g" % (poresize, K), fontsize=20)

        plot_curves(ax1, t, t0, 0, "Theoretical Time Axis", timescale=1)

        tI = -500
        x = t + tI
        x0 = t0 + tI
        plot_curves(ax2, x, x0, tI, "Experimental Time Axis", timescale=0.5)

        fig.tight_layout()

        ret = plt.show()
    return ret


"""
    RtEmgParams.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from .EghParams import EghParamsBase, get_common_parameter_names
from molass_legacy._MOLASS.SerialSettings import get_setting
from SecTheory.RetensionTime import compute_retention_time
from SecTheory.HermansEmg import convert_to_xr_params_hermans
from SecTheory.ColumnConstants import SECCONF_LOWER_BOUND, BAD_CONFORMANCE_REDUCE
from .SimpleSecParams import SEC_PENALTY_SCALE

SEC_START = -5
SEC_STOP = -1

class RtEmgParams(EghParamsBase):
    def __init__(self, n_components):
        EghParamsBase.__init__(self, n_components, None, None)
        self.advanced = True
        nc = n_components - 1

        self.pos = []
        self.pos.append(0)      # [0] xr_params
        sep = nc
        self.pos.append(sep)    # [1] xr_baseparams
        sep += self.num_baseparams
        self.pos.append(sep)    # [2] rgs
        sep += nc
        self.pos.append(sep)    # [3] mapping
        sep = sep+2
        self.pos.append(sep)    # [4] uv_params
        sep += nc
        self.pos.append(sep)    # [5] uv_baseparams
        sep += 5 + self.num_baseparams
        self.pos.append(sep)    # [6] mappable_range
        sep += 2
        self.pos.append(sep)    # [7] end

        self.num_params = sep   # note that this does not include sec params

    def __str__(self):
        return "RtEmgParams(nc=%d)" % (self.n_components - 1)

    def get_estimator(self, editor, debug=True):
        if debug:
            from importlib import reload
            import Estimators.RtEmgEstimator
            reload(Estimators.RtEmgEstimator)
        from Estimators.RtEmgEstimator import RtEmgEstimator
        return RtEmgEstimator(editor)

    def split_params(self, n, params):
        if self.num_params != len(params):
            raise ValueError("len(params)=%d != %d, which is calculated from number of components %d." % (len(params), self.num_params, n))

        ret_params = []
        for p, q in zip(self.pos[:-1], self.pos[1:]):
            ret_params.append(params[p:q])

        return ret_params

    def split_params_simple(self, params, convert=True):
        req_params = params[0:SEC_START]
        sec_params = params[SEC_START:SEC_STOP]
        R = params[SEC_STOP]
        split_params = self.split_params(self.n_components, req_params)
        if convert:
            xr_params = convert_to_xr_params_hermans(split_params[0], split_params[2], sec_params, R)
        else:
            xr_params = split_params[0]
        self.separate_params = [xr_params] + split_params[1:] + [sec_params, R]
        return self.separate_params

    def split_as_unified_params(self, params, **kwargs):
        class NotSupportedError(Exception): pass
        raise NotSupportedError()

    def get_param_bounds(self, params):

        init_xr_params, init_xr_baseparams, init_rgs, init_mapping, init_uv_params, init_uv_baseparams, init_mappable_range = self.split_params_simple(params)[0:7]

        xr_h_max = np.max(init_xr_params[:])
        m_allow = 100

        xr_bounds = []
        for h in init_xr_params:
            xr_bounds.append((0, xr_h_max*2))       # h

        for k, v in enumerate(init_xr_baseparams):
            v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
            if self.integral_baseline and k == 2:
                bounds = (max(0, v - v_allow), v + v_allow)
            else:
                bounds = (v - v_allow, v + v_allow)
            xr_bounds.append(bounds)

        rg_bounds = [(rg*0.5, rg*2) for rg in init_rgs]
        a, b = init_mapping
        mapping_bounds = [(a*0.8, a*1.2), (-m_allow, m_allow)]

        uv_h_max = np.max(init_uv_params)
        uv_bounds = [(0, uv_h_max*2) for h in init_uv_params]
        for k, v in enumerate(init_uv_baseparams):
            v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
            if self.integral_baseline and k == 7:
                bounds = (max(0, v - v_allow), v + v_allow)
            else:
                bounds = (v - v_allow, v + v_allow)
            uv_bounds.append(bounds)

        f, t = init_mappable_range
        dx = (t - f)*0.1
        range_bounds = [(f-dx, f+dx), (t-dx, t+dx)]

        """
        task: unify bounds setting
        """

        Np = get_setting("numplates_pc")
        RATE_R_UPPER_BOUND = get_setting("RATE_R_UPPER_BOUND")

        #               t0,         P,          rp,         m,      R
        sec_bounds = [(-200, 200), (200, 3000), (30, 150), (1, 6), (0, RATE_R_UPPER_BOUND)]

        self.bounds_lengths = [len(b) for b in [xr_bounds, rg_bounds, mapping_bounds, uv_bounds, range_bounds, sec_bounds]]
        return xr_bounds + rg_bounds + mapping_bounds + uv_bounds + range_bounds + sec_bounds

    def make_bounds_mask(self):
        bounds_mask = np.zeros(self.num_params + 5, dtype=bool)
        nc = self.n_components - 1
        xr_base_start = nc
        bounds_mask[0:xr_base_start] = True

        rg_start = xr_base_start + self.num_baseparams

        if self.integral_baseline:
            bounds_mask[rg_start-1] = True      # xr baseline fouling
            bounds_mask[self.pos[6]-1] = True   # uv baseline fouling

        bounds_mask[rg_start:rg_start+nc] = True
        return bounds_mask

    def update_bounds_hook(self, masked_init_params):
        # nothing to do
        pass

    def get_rg_start_index(self):
        return self.pos[2]

    def get_mr_start_index(self):
        return self.pos[6]

    def get_xr_parameter_names(self):
        nc = self.n_components - 1
        xr_names = []
        for k in range(nc):
            xr_names += ["$h_%d$" % k]
        return xr_names

    def get_parameter_names(self):
        n = self.n_components
        _, rg_names, mapping_names, uv_names, mr_names, seccol_names = get_common_parameter_names(n, nc_rg=n-1)
        xr_names = self.get_xr_parameter_names()
        xr_basenames = ["$xb_a$", "$xb_b$"]
        if self.num_baseparams == 3:
            xr_basenames += ["$xb_r$"]
        uv_basenames = ["$L$", "$x_0$", "$k$", "$b$", "$s_1$", "$s_2$", "$diffratio$"]
        if self.num_baseparams == 3:
            uv_basenames += ["$ub_r$"]
        return np.array(xr_names + xr_basenames + rg_names + mapping_names + uv_names + uv_basenames + mr_names + seccol_names + ["$R$"])

    def get_peak_pos_array_list(self, x_array):
        ni = x_array.shape[0]
        nc = self.n_components - 1
        ret_pos_array = np.zeros((nc, ni))
        xr_start = 0
        xr_stop = nc
        rg_start = nc + 2
        rg_stop = nc + 2 + nc
        for i, p in enumerate(x_array):
            xr_params = convert_to_xr_params_hermans(p[xr_start:xr_stop], p[rg_start:rg_stop], p[SEC_START:SEC_STOP], p[SEC_STOP])
            for k in range(nc):
                ret_pos_array[k,i] = xr_params[k,1] + xr_params[k,2]    # i.e., mu + tau
        return ret_pos_array

    def split_get_unified_sec_params(self, params):
        separate_params = self.split_params_simple(params)
        t0, K, rp, m = separate_params[-2]
        R = separate_params[-1]
        return t0, K, rp, m, None, None, None,None, None, None, R

    def compute_xr_peak_position(self, i, xr_params):
        return xr_params[i,1] + xr_params[i,3]



# coding: utf-8
"""
    QmmMenu.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from MenuButton import MenuButton

class QmmMenu(Tk.Frame):
    def __init__(self, parent, dialog, in_folder):
        Tk.Frame.__init__(self, parent)
        self.parent = parent
        self.dialog = dialog
        self.in_folder = in_folder
        self.menu = MenuButton(self, "▶▶ QMM Menu", [
                            ("EGH-QMM", self.show_egh_qmm_dialog),
                            ("EMG-QMM", self.show_emg_qmm_dialog),
                            ("QMM Options", self.show_qmm_options_dialog),
                            ("QMM Window", self.show_qmm_window_setting_dialog),
                            ])
        self.menu.pack()

    def prepare_denoise_rank(self, dialog):
        """
        called from dialog(GuiMain) when ready
        """
        from .PairedDataSets import get_denoise_rank_impl
        from molass_legacy._MOLASS.SerialSettings import set_setting
        sd = dialog.serial_data
        denoise_rank = get_denoise_rank_impl(sd.get_xray_curve())
        set_setting('last_denoise_rank', denoise_rank)

    def config(self, **kwargs):
        self.menu.config(**kwargs)

    def show_egh_qmm_dialog(self):
        from Prob.EghMixture import EghMixture
        self.show_qmm_dialog(EghMixture)

    def show_emg_qmm_dialog(self):
        from Prob.EmgMixture import EmgMixture
        self.show_qmm_dialog(EmgMixture)

    def show_qmm_dialog(self, mm_type):
        from Prob.QmmController import get_qmm_controller
        from molass_legacy.SerialAnalyzer.AbnormalityCheck import update_abnormality_fix_state

        dialog = self.dialog

        if not dialog.get_it_ready_for_qmm():
            return

        self.qmm_controller = None
        try:
            update_abnormality_fix_state(dialog.serial_data, dialog.file_info_table, dialog.analyzer.app_logger)

            in_folder = self.in_folder.get()
            dialog.is_executing = True
            dialog.qmm_controller = get_qmm_controller(dialog, mm_type, in_folder, sd=dialog.serial_data)
            assert dialog.qmm_controller is not None
            dialog.is_executing = False

            ret = dialog.qmm_controller.show_dialog(dialog)
        except:
            try:
                import molass_legacy.KekLib.CustomMessageBox         as MessageBox
            except:
                import OurMessageBox            as MessageBox
            if dialog.testing:
                self.dialog.quit(immediately=True)
            else:
                import sys
                message = "error occured in QmmDialog\n" + str(sys.exc_info())
                MessageBox.showerror( "QmmDialog Error", message, parent=dialog.parent )
            ret = False

        if ret:
            print("to Serial Analysis")

    def show_qmm_options_dialog(self):
        from .QmmOptions import QmmOptionsDialog
        opts_dialog = QmmOptionsDialog(self.dialog)
        opts_dialog.show()

    def show_qmm_window_setting_dialog(self):
        from .QmmWindowSetting import QmmWindowSetting
        window_dialog = QmmWindowSetting(self.dialog, self.dialog.fig_frame)
        window_dialog.show()

class MockCurve:
    def get_major_peak_info(self):
        return [None]*3

class MockSd:
    def get_xray_curve(self):
        return MockCurve()

class MockAnalyzer:
    def __init__(self):
        import logging
        self.app_logger = logging.getLogger(__name__)

class App(Dialog):
    def __init__(self, parent, in_folder):
        self.in_folder = Tk.StringVar()
        self.in_folder.set(in_folder)
        self.testing = False
        self.serial_data = MockSd()
        self.fig_frame = None
        self.analyzer = MockAnalyzer()
        Dialog.__init__(self, parent, location="lower right")

    def body(self, body_frame):
        label = Tk.Label(body_frame, textvariable=self.in_folder)
        label.pack()
        menu = QmmMenu(body_frame, self, self.in_folder)
        menu.pack()

    def get_it_ready_for_qmm(self):
        from molass_legacy._MOLASS.SerialSettings import set_setting
        set_setting( 'analysis_name', 'analysis-000' )
        return True



import sys
import os
import numpy    as np
import matplotlib.pyplot    as plt
from scipy.optimize import fsolve

def f_numerator( xR ):
    return np.sin( xR ) - xR*np.cos( xR )

if len(sys.argv) < 3:
    R = 23
    z = 0.2
else:
    R = float(sys.argv[1])
    z = float(sys.argv[2])

print( 'R=', R, 'z=', z )

zeros_init = np.array( [z] )
x0R = fsolve(f_numerator, zeros_init*R )
x0 = x0R/R
# print( 'x0R=', x0R, x0, f_numerator(x0R) )

print( 'x0=', x0 )

x = np.linspace( 0.005, 0.6, 800 )

plt.plot( x, f_numerator( x*R ) )
plt.plot( x0, np.zeros( len(x0) ), 'or' )
plt.show()



"""
    ReconsiderDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import numpy as np
from bisect import bisect_right
from scipy.stats import linregress
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from Theory.SolidSphere import phi

def demo_impl(sd, f, t):
    from DataUtils import get_in_folder
    from importlib import reload
    import BoundedLRF.NaiiveLrfSolver
    reload(BoundedLRF.NaiiveLrfSolver)
    from .NaiiveLrfSolver import NaiiveLrfSolver
    import BoundedLRF.BoundedLrfSolver
    reload(BoundedLRF.BoundedLrfSolver)
    from .BoundedLrfSolver import BoundedLrfSolver

    print("demo_impl")

    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    x = xr_curve.x
    y = xr_curve.y
    data_title = get_in_folder()

    i = bisect_right(qv, 0.02)
    cy = D[i,:]
    j = np.argmax(cy)

    print("len(cy)=", len(cy), "j=", j)
    pv1 = D[:,j]
    # pv1 = pv1/np.max(pv1)

    range_ = slice(f, t+1)

    solver = NaiiveLrfSolver(qv, D[:,range_], E[:,range_])
    P, C = solver.solve()
    c1, c2 = C[:,j-f]
    aq, bq = P.T
    aq1 = aq*c1
    bq1 = bq

    range_ = slice(f,t+1)
    debug_info = sd, range_
    solver = BoundedLrfSolver(qv, D[:,range_], E[:,range_], debug_info=debug_info)
    P_, C_, Rg, R_, L_, hK, hL, bq_bounds_, coerced_bq_ = solver.solve(debug=True)
    aq_, bq_ = P_.T
    aqf = aq_*c1

    print("L_=", L_)

    TxT = True

    with plt.Dp():
        if TxT:
            fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(21,10))
            ax0 = axes[0,0]
            ax1 = axes[0,1]
            ax2 = axes[0,2]
            ax3 = axes[1,1]
            ax4 = axes[1,0]
        else:
            fig, (ax0, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(22,5))
        fig.suptitle("Bounded LRF Demo for %s" % (data_title), fontsize=20)
        ax0.set_title("Elution Data and Range", fontsize=16)
        ax2.set_title("LRF Result (Linear)", fontsize=16)
        ax3.set_title("LRF Result (Log)", fontsize=16)
        ax4.set_title("LRF Result (Guinier Plot)", fontsize=16)

        ax0.plot(x, y)
        ymin, ymax = ax0.get_ylim()
        p = Rectangle(
                (f, ymin),  # (x,y)
                t - f,   # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax0.add_patch(p)

        

        ax2.plot(qv, bq_*c2, color="pink", label="bounded LRF B(q)")
        if False:
            h_bq = -hK*phi(qv, 2*hL*R_)*aqf/c1
            ax2.plot(qv, h_bq*c2, color="yellow", label="Hard Sphere fit B(q)")

        ymin, ymax = ax2.get_ylim()
        ax2.set_ylim(ymin, ymax)
        for k, bound in enumerate(bq_bounds_):
            label = "estimated B(q) bound" if k == 0 else None
            ax2.plot(qv, bound*c2, color="red", alpha=0.5, label=label)

        ax2.legend(fontsize=14)

        ax3.set_yscale("log")
        for ax in ax2, ax3:
            ax.plot(qv, pv1, label="measured data")
            ax.plot(qv, aq1, label="naiive LRF A(q)", alpha=0.5)
            ax.plot(qv, aqf, label="bounded LRF A(q)", alpha=0.5)
        ax3.legend(fontsize=14)

        glim_i = bisect_right(qv, 2/Rg)
        gslice_disp = slice(0,glim_i)
        qv2 = qv[gslice_disp]**2

        glim_i = bisect_right(qv, 1.3/Rg)
        gslice = slice(0,glim_i)
        qv2_ = qv[gslice]**2
        intercept_list = []
        rgs = []
        for pv in pv1, aq1, aqf:
            glny_ = np.log(pv[gslice])
            slope, intercept = linregress(qv2_, glny_)[0:2]
            Rg = np.sqrt(-3*slope)
            rgs.append(Rg)
            # I0 = np.exp(intercept)
            intercept_list.append(intercept)

        ax4.plot(qv2, np.log(pv1[gslice_disp]) - intercept_list[0], label="measured data")
        ax4.plot(qv2, np.log(aq1[gslice_disp]) - intercept_list[1], label="naiive LRF A(q)", alpha=0.5)
        ax4.plot(qv2, np.log(aqf[gslice_disp]) - intercept_list[2], label="bounded LRF A(q)", alpha=0.5)
        ymin, ymax = ax4.get_ylim()
        ax4.set_ylim(ymin, ymax)
        qg = qv[glim_i]**2
        ax4.plot([qg, qg], [ymin, ymax], color="yellow", label="$qR_g=1.3$")
        xmin, xmax = ax4.get_xlim()
        tx = xmin*0.8 + xmax*0.2
        ty = ymin*0.7 + ymax*0.3
        ax4.text(tx, ty, "$R_g=%.3g$" % rgs[-1], ha="center", va="center", alpha=0.5, fontsize=20)
        ax4.legend(fontsize=14)

        fig.tight_layout()
        plt.show()

def demo(caller):
    dialog = caller.dialog
    pdata, popts = caller.get_preview_data(with_update=False)

    print("pdata.cnv_ranges=", pdata.cnv_ranges)
    paired_range = pdata.cnv_ranges[1]
    ranges = paired_range.get_fromto_list()
    print(ranges)
    f, t = ranges[1]
    sd = dialog.sd
    demo_impl(sd, f, t)

def demo_bak2(caller):
    from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
    from molass_legacy.Tools.EmbedCushionUtils import get_caller_attr

    in_folder_save = get_setting("in_folder")

    gi_in_folder = r"E:\PyTools\Data\20180526\GI"
    if not os.path.exists(gi_in_folder):
        gi_in_folder = r"D:\PyTools\Data\20180526\GI"
    cushion_dict = get_caller_attr(caller, "cushion_dict", {})
    gi_sd = cushion_dict.get("gi_sd")
    if gi_sd is None:
        from molass_legacy.Tools.SdUtils import get_sd
        cushion_dict["gi_sd"] = gi_sd = get_sd(gi_in_folder)

    set_setting("in_folder", gi_in_folder)

    f = 170
    t = 200
    demo_impl(gi_sd, f, t)

    set_setting("in_folder", in_folder_save)

def demo_bak3(caller):
    from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
    from molass_legacy.Tools.EmbedCushionUtils import get_caller_attr

    in_folder_save = get_setting("in_folder")

    gi_in_folder = r"E:\PyTools\Data\20180526\OA"
    if not os.path.exists(gi_in_folder):
        gi_in_folder = r"D:\PyTools\Data\20180526\OA"
    cushion_dict = get_caller_attr(caller, "cushion_dict", {})
    gi_sd = cushion_dict.get("oa_sd")
    if gi_sd is None:
        from molass_legacy.Tools.SdUtils import get_sd
        cushion_dict["oa_sd"] = gi_sd = get_sd(gi_in_folder)

    set_setting("in_folder", gi_in_folder)

    f = 250
    t = 310
    demo_impl(gi_sd, f, t)

    set_setting("in_folder", in_folder_save)


def demo_bak4(caller):
    from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
    from molass_legacy.Tools.EmbedCushionUtils import get_caller_attr

    in_folder_save = get_setting("in_folder")

    temp_in_folder = r"E:\PyTools\Data\20221227"
    if not os.path.exists(temp_in_folder):
        temp_in_folder = r"D:\PyTools\Data\20221227"
    cushion_dict = get_caller_attr(caller, "cushion_dict", {})
    temp_sd = cushion_dict.get("temp_sd")
    if temp_sd is None:
        from molass_legacy.Tools.SdUtils import get_sd
        cushion_dict["temp_sd"] = temp_sd = get_sd(temp_in_folder)

    set_setting("in_folder", temp_in_folder)

    f = 220
    t = 260
    demo_impl(temp_sd, f, t)

    set_setting("in_folder", in_folder_save)



# coding: utf-8
"""
    OurOptics.py

    custom class of opticspy

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm as cm
from opticspy.zernike import Coefficient
from opticspy import interferometer_zenike as __interferometer__

class ZernikeCoefficient(Coefficient):
    def __init__(self, **kwargs):
        Coefficient.__init__(self, **kwargs)

    def zernikemap(self, fig=None, ax=None, cmap=None, label=True):

        theta = np.linspace(0, 2*np.pi, 400)
        rho = np.linspace(0, 1, 400)
        [u,r] = np.meshgrid(theta,rho)
        X = r*np.cos(u)
        Y = r*np.sin(u)
        Z = __interferometer__.__zernikepolar__(self.__coefficients__,r,u)

        arg_ax = ax
        if arg_ax is None:
            fig = plt.figure(figsize=(12, 8), dpi=80)
            ax = fig.gca()

        if cmap is None:
            cmap = cm.RdYlGn
        im = ax.pcolormesh(X, Y, Z, cmap=cmap)

        if label:
            plt.title('Zernike Polynomials Surface Heat Map',fontsize=18)
            ax.set_xlabel(self.listcoefficient()[1],fontsize=18)

        fig.colorbar(im, ax=ax)
        ax.set_aspect('equal', 'datalim')

        if arg_ax is None:
            plt.show()
